# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import difflib
import pprint
import re
import sys
import traceback


MAX_REPR_LENGTH = 1024*100
DIFF_THRESHOLD = 1024*100


def set_diff_threshold(threshold):
    """
    Set the diff threshold to ``threshold``:

    Args:
        threshold (int): the maximum number of characters to use difflib
    """
    global DIFF_THRESHOLD
    assert isinstance(threshold, int), "threshold must be an int"
    DIFF_THRESHOLD = threshold


def set_repr_length(length):
    """
    Set the maximum representation length to ``length``

    Args:
        length (int): the maximum representation length
    """
    global MAX_REPR_LENGTH
    assert isinstance(length, int), "length must be an int"
    MAX_REPR_LENGTH = length


def _safe_representation(obj):
    try:
        result = repr(obj)
    except Exception:
        result = object.__repr__(obj)

    if len(result) < MAX_REPR_LENGTH:
        return result

    return result[:MAX_REPR_LENGTH] + ' (>% >% >% truncated to {} characters) ...'.format(MAX_REPR_LENGTH)


def _format_message(message=None, default_message=None):
    if message is None:
        return default_message
    else:
        if default_message is not None:
            return message + ": " + default_message
        else:
            return message


def assert_true(expression, message=None):
    """
    Assert that ``expression`` is ``True``.

    Args:
        expression (bool): the boolean expression to be validated
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``expression`` is not ``True``
    """
    if not expression:
        message = _format_message(message, "{} is not true".format(_safe_representation(expression)))
        raise AssertionError(message)


def assert_false(expression, message=None):
    """
    Assert that ``expression`` is ``False``.

    Args:
        expression (bool): the boolean expression to be validated
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``expression`` is not ``False``
    """
    if expression:
        message = _format_message(message, "{} is not false".format(_safe_representation(expression)))
        raise AssertionError(message)


_type_equality_functions = {}


def set_equality_asserter(type_, asserter):
    """
    Register a new type equality asserter used in :func:`~htf.assert_equal` and :func:`~htf.assert_not_equal`.

    Args:
        type_ (type): the type the checker is used for
        asserter (callable): callable equality asserter object that supports parameters ``a``, ``b`` and ``message``
    """
    _type_equality_functions[type_] = asserter


def get_equality_asserter(type_):
    """
    Return the registered equality asserter for ``type_``.

    Args:
        type_ (type): a type

    Returns:
        the registered equality asserter for ``type``

    Raises:
        KeyError: if ``type_`` has no registered equality asserter
    """
    return _type_equality_functions[type_]


def _get_equality_asserter(a, b):
    """
    Get a detailed comparison asserter for the needed type.

    Args:
        a: an object
        b: an object

    Returns:
        The equality asserter callable for to compare ``a`` and ``b`` or the base asserter if no asserter was found or
            the types of ``a`` and ``b`` differ.
    """
    if type(a) is type(b):
        return _type_equality_functions.get(type(a), _base_assert_equal)
    else:
        return _base_assert_equal


def assert_sequence_equal(a, b, message=None, sequence_type=None):
    """
    Assert that ``a`` equals ``b`` and that both are sequences.

    Args:
        a: a sequence
        b: a sequence
        sequence_type=None (type): the sequence type to be used (if not ``None``)
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not equal or don't match the ``sequence_type``
    """
    if sequence_type is not None:
        sequence_type_name = sequence_type.__name__

        if not isinstance(a, sequence_type):
            raise AssertionError(
                "'a' is not a {}: {}".format(sequence_type_name, a)
            )

        if not isinstance(b, sequence_type):
            raise AssertionError(
                "'b' is not a {}: {}".format(sequence_type_name, b)
            )
    else:
        sequence_type_name = "sequence"

    differences = None

    try:
        len_a = len(a)
    except (TypeError, NotImplementedError):
        differences = "{} 'a' has no length (len()). Is 'a' a {}?".format(sequence_type_name, sequence_type_name)

    if differences is None:
        try:
            len_b = len(b)
        except (TypeError, NotImplementedError):
            differences = "{} 'b' has no length (len()). Is 'b' a {}?".format(sequence_type_name, sequence_type_name)

    if differences is None:
        if a == b:
            return

        differences = "{}s 'a' and 'b' differ\n".format(sequence_type_name)

        for i in range(min(len_a, len_b)):
            try:
                item1 = a[i]
            except (TypeError, IndexError, NotImplementedError):
                differences += "\nUnable to index element {} of 'a'\n".format((i))
                break

            try:
                item2 = b[i]
            except (TypeError, IndexError, NotImplementedError):
                differences += "\nUnable to index element {} of 'b'\n".format(i)
                break

            if item1 != item2:
                differences += "\nFirst differing element {}:\n{}\n{}\n".format(i, item1, item2)
                break
        else:
            if len_a == len_b and sequence_type is None and type(a) != type(b):
                # a and b have the same contents and different types
                return

        if len_a > len_b:
            differences += "\n'a' contains {} additional elements.\n".format(len_a - len_b)

            try:
                differences += "First extra element {}:\n{}\n".format(len_b, _safe_representation(a[len_b]))
            except (TypeError, IndexError, NotImplementedError):  # pragma: no cover
                differences += "Unable to index element {} of 'a' {}\n".format(len_b, sequence_type_name)
        elif len_a < len_b:
            differences += "\n'b' contains {} additional elements.\n".format(len_b - len_a)

            try:
                differences += "First extra element {}:\n{}\n".format(len_a, _safe_representation(b[len_a]))
            except (TypeError, IndexError, NotImplementedError):  # pragma: no cover
                differences += "Unable to index element {} of 'b'' {}\n".format(len_a, sequence_type_name)

    default_message = differences

    default_message += "\n" + "\n".join(
        difflib.ndiff(pprint.pformat(a).splitlines(), pprint.pformat(b).splitlines()))

    raise AssertionError(
        _format_message(message, default_message)
    )


def assert_list_equal(a, b, message=None):
    """
    Assert that ``a`` equals ``b`` and that both are of type ``list``.

    Args:
        a: a list
        b: a list
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not equal or not lists
    """
    assert_sequence_equal(a, b, message=message, sequence_type=list)


def assert_tuple_equal(a, b, message=None):
    """
    Assert that ``a`` equals ``b`` and that both are of type ``tuple``.

    Args:
        a: a list
        b: a list
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not equal or not tuples
    """
    assert_sequence_equal(a, b, message=message, sequence_type=tuple)


def assert_set_equal(a, b, message=None):
    """
    Assert that ``a`` equals ``b`` and that both are of type ``set``.

    Args:
        a: a set
        b: a set
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not equal or not sets
    """
    try:
        difference_1 = a.difference(b)
    except TypeError as e:
        raise TypeError("invalid type when attempting set difference: {}".format(e))
    except AttributeError as e:
        raise TypeError("a does not support set difference: {}".format(e))

    try:
        difference_2 = b.difference(a)
    except TypeError as e:
        raise TypeError("invalid type when attempting set difference: {}".format(e))
    except AttributeError as e:
        raise TypeError("b does not support set difference: {}".format(e))

    if not (difference_1 or difference_2):
        return

    lines = []

    if difference_1:
        lines.append('Items in set a but not in set b:')

        for item in difference_1:
            lines.append(repr(item))

    if difference_2:
        lines.append('Items in set b but not in set a:')

        for item in difference_2:
            lines.append(repr(item))

    default_message = '\n'.join(lines)

    raise AssertionError(
        _format_message(message, default_message)
    )


def assert_dict_equal(a, b, message=None):
    """
    Assert that ``a`` equals ``b`` and that both are of type ``dict``.

    Args:
        a: a dict
        b: a dict
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not equal
    """
    assert_is_instance(a, dict, 'a argument is not a dictionary')
    assert_is_instance(b, dict, 'b argument is not a dictionary')

    if a != b:
        default_message = "{} != {}".format(a, b)
        raise AssertionError(
            _format_message(message, default_message)
        )


def assert_multi_line_equal(a, b, message=None):
    """
    Assert that ``a`` equals ``b`` and that both are of type ``str``.

    Args:
        a: a string
        b: a string
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not equal or not strings
    """
    assert_is_instance(a, type(""), 'First argument is not a string')
    assert_is_instance(b, type(""), 'Second argument is not a string')

    if not a == b:
        # do not use difflib if a or b is longer than the threshold
        if len(a) > DIFF_THRESHOLD or len(b) > DIFF_THRESHOLD:
            assert_sequence_equal(a, b, message=message, sequence_type=type(""))
        else:
            a_lines = a.splitlines(True)
            b_lines = b.splitlines(True)

            if len(a_lines) == 1 and a.strip('\r\n') == a:
                a_lines = [a + '\n']
                b_lines = [b + '\n']

            default_message = "strs 'a' and 'b' differ\n".format(a, b)

            default_message += "".join(difflib.ndiff(a_lines, b_lines))

            raise AssertionError(
                _format_message(message, default_message)
            )


assert_str_equal = assert_multi_line_equal

set_equality_asserter(dict, assert_dict_equal)
set_equality_asserter(list, assert_list_equal)
set_equality_asserter(tuple, assert_tuple_equal)
set_equality_asserter(set, assert_set_equal)
set_equality_asserter(frozenset, assert_set_equal)
set_equality_asserter(str, assert_multi_line_equal)

def _base_assert_equal(a, b, message=None):
    """
    Assert that ``a`` equals ``b``.

    Args:
        a: an object
        b: an object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not equal
    """
    if not a == b:
        default_message = "{} != {}".format(
            _safe_representation(a),
            _safe_representation(b)
        )

        raise AssertionError(
            _format_message(message, default_message)
        )


def assert_equal(a, b, message=None):
    """
    Assert that ``a`` equals ``b``. Tries to find a registered equality asserter and uses the base assert if
    no special assert is registered.

    Args:
        a: an object
        b: an object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not equal
    """
    asserter = _get_equality_asserter(a, b)
    asserter(a, b, message=message)


def assert_not_equal(a, b, message=None):
    """
    Assert that ``a`` does not equal ``b``.

    Args:
        a: an object
        b: an object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are equal
    """
    if a == b:
        default_message = "{} == {}".format(
            _safe_representation(a),
            _safe_representation(b)
        )

        raise AssertionError(
            _format_message(message, default_message)
        )


def assert_almost_equal(a, b, places=None, delta=None, message=None):
    """
    Assert that ``a`` and ``b`` are almost equal. Useful to compare floating point numbers.
    Comparison is realized either using a maximum ``delta`` or the number of decimal ``places`` 

    Args:
        a (float): number
        b (float): number
        places=None (int): the number of decimal places for comparison. If ``None`` ``7`` decimal places are used.
        delta=None (float): the maximum delta
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are not almost equal
    """
    if delta is not None and places is not None:
        raise TypeError("specify delta or places but not both")

    if a == b:
        # a and b are equal
        return

    difference = abs(a - b)

    if delta is not None:
        # check by delta
        if difference <= delta:
            return

        default_message = "{} != {} within {} delta ({} difference)".format(
            _safe_representation(a),
            _safe_representation(b),
            _safe_representation(delta),
            _safe_representation(difference))
    else:
        # check by places
        if places is None:
            places = 7

        if round(difference, places) == 0:
            return

        default_message = "{} != {} within {} places ({} difference)".format(
            _safe_representation(a),
            _safe_representation(b),
            places,
            _safe_representation(difference))

    raise AssertionError(
        _format_message(message, default_message)
    )


def assert_not_almost_equal(a, b, places=None, delta=None, message=None):
    """
    Assert that ``a`` and ``b`` are not almost equal. Useful to compare floating point numbers.
    Comparison is realized either using a maximum ``delta`` or the number of decimal ``places``

    Args:
        a (float): number
        b (float): number
        places=None (int): the number of decimal places for comparison. If ``None`` ``7`` decimal places are used.
        delta=None (float): the maximum delta
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` and ``b`` are almost equal
    """

    if delta is not None and places is not None:
        raise TypeError("specify delta or places but not both")

    diff = abs(a - b)

    if delta is not None:
        if not a == b and diff > delta:
            return

        default_message = "{} == {} within {} delta ({} difference)".format(
            _safe_representation(a),
            _safe_representation(b),
            _safe_representation(delta),
            _safe_representation(diff))
    else:
        if places is None:
            places = 7
        if not a == b and round(diff, places) != 0:
            return
        default_message = "{} == {} within {} places".format(
            _safe_representation(a),
            _safe_representation(b),
            places)

    raise AssertionError(
        _format_message(message, default_message)
    )


class _AssertRaisesContext:
    def __init__(self, expected_exception, expected_regex=None, callable_object=None):
        self._expected_exception = expected_exception
        self._expected_regex = expected_regex
        self._callable_object = callable_object

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, tb):
        if exception_type is None:
            # no exception occurred
            try:
                exception_name = self._expected_exception.__name__
            except AttributeError:  # pragma: no cover
                exception_name = str(self._expected_exception)

            raise AssertionError(
                _format_message(
                    "{} was not raised".format(exception_name)
                )
            )

        traceback.clear_frames(tb)

        if not issubclass(exception_type, self._expected_exception):
            # re-raise unexpected exceptions as is
            return False

        # make exception available for use with the context manager
        self.exception = exception_value.with_traceback(None)

        if self._expected_regex is None:
            return True

        # handle regular expression
        expected_regex = re.compile(self._expected_regex)

        if not expected_regex.search(str(exception_value)):
            raise AssertionError(
                _format_message(
                    "'{}' does not match '{}'".format(
                        expected_regex.pattern,
                        str(exception_value)
                    )
                )
            )

        return True


def assert_raises(expected_exception, callable_object=None, *args, **kwargs):
    """
    Assert that ``expected_exception`` is raised. Can be used as a context manager or
    with a callable object that is run with ``*args`` and ``**kwargs``.

    Args:
        expected_exception (type): the expected exception that shall be raised
        callable_object=None (callable): the callable object if not used as a context manager
        *args: variable positional arguments passed to ``callable_object``
        **kwargs: variable keyword arguments passed to ``callable_object``

    Returns:
        a context manager if ``callable_object`` is ``None`` else ``None``.

    Raises:
        AssertionError: if ``expected_exception`` is not raised
    """
    context = _AssertRaisesContext(expected_exception)

    if callable_object is None:
        return context

    with context:
        callable_object(*args, **kwargs)


def assert_raises_regex(expected_exception, expected_regex, callable_object=None, *args, **kwargs):
    """
    Assert that ``expected_exception`` is raised. Can be used as a context manager or
    with a callable object that is run with ``*args`` and ``**kwargs``.

    Args:
        expected_exception (type): the expected exception that shall be raised
        expected_regex (regular expression): the regular expression that shall match the stringified representation
            of the exception
        callable_object=None (callable): the callable object if not used as a context manager
        *args: variable positional arguments passed to ``callable_object``
        **kwargs: variable keyword arguments passed to ``callable_object``

    Returns:
        a context manager if ``callable_object`` is ``None`` else ``None``.

    Raises:
        AssertionError: if ``expected_exception`` is not raised or does not match ``expected_regex``
    """
    context = _AssertRaisesContext(expected_exception, expected_regex=expected_regex)

    if callable_object is None:
        return context

    with context:
        callable_object(*args, **kwargs)


def assert_in(member, container, message=None):
    """
    Assert that ``container`` contains ``member``.

    Args:
        member: the object that shall be included in ``container``
        container (iterable): an iterable to search for ``member``
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``member`` is not included in ``container``
    """
    if member not in container:
        raise AssertionError(
            _format_message(
                message, "{} not found in {}".format(member, container)
            )
        )


def assert_not_in(member, container, message=None):
    """
    Assert that ``container`` does not contain ``member``.

    Args:
        member: the object that shall not be included in ``container``
        container (iterable): an iterable to search for ``member``
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``member`` is included in ``container``
    """
    if member in container:
        raise AssertionError(
            _format_message(
                message, "{} unexpectedly found in {}".format(member, container)
            )
        )


def assert_less(a, b, message=None):
    """
    Assert that ``a`` is less than ``b``.

    Args:
        a: a comparable object
        b: a comparable object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` is not less than ``b``
    """
    if not a < b:
        raise AssertionError(
            _format_message(
                message, "{} is not less than {}".format(a, b)
            )
        )


def assert_less_equal(a, b, message=None):
    """
    Assert that ``a`` is less or equal to ``b``.

    Args:
        a: a comparable object
        b: a comparable object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` is not less or equal to ``b``
    """
    if not a <= b:
        raise AssertionError(
            _format_message(
                message, "{} is not less or equal to {}".format(a, b)
            )
        )


def assert_greater(a, b, message=None):
    """
    Assert that ``a`` is greater than ``b``.

    Args:
        a: a comparable object
        b: a comparable object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` is not greater than ``b``
    """
    if not a > b:
        raise AssertionError(
            _format_message(
                message, "{} is not greater than {}".format(a, b)
            )
        )


def assert_greater_equal(a, b, message=None):
    """
    Assert that ``a`` is greater or equal to ``b``.

    Args:
        a: a comparable object
        b: a comparable object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``a`` is not greater or equal to ``b``
    """
    if not a >= b:
        raise AssertionError(
            _format_message(
                message, "{} is not greater or equal to {}".format(a, b)
            )
        )


def assert_is(a, b, message=None):
    """
    Assert that ``id(a)`` equals ``id(b)``.

    Args:
        a: an object
        b: another object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``id(a)`` does not equal ``id(b)``
    """
    if a is not b:
        raise AssertionError(
            _format_message(
                message, "{} is not {}".format(a, b)
            )
        )


def assert_is_not(a, b, message=None):
    """
    Assert that ``id(a)`` does not equal ``id(b)``.

    Args:
        a: an object
        b: another object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``id(a)`` equals ``id(b)``
    """
    if a is b:
        raise AssertionError(
            _format_message(
                message, "{} is not {}".format(a, b)
            )
        )


def assert_is_none(obj, message=None):
    """
    Assert that ``obj`` is ``None``.

    Args:
        obj: an object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``obj`` is not ``None``
    """
    if obj is not None:
        raise AssertionError(
            _format_message(
                message, "{} is not None".format(obj)
            )
        )


def assert_is_not_none(obj, message=None):
    """
    Assert that ``obj`` is not ``None``.

    Args:
        obj: an object
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``obj`` is  ``None``
    """
    if obj is None:
        raise AssertionError(
            _format_message(
                message, "{} is None".format(obj)
            )
        )


def assert_is_instance(obj, cls, message=None):
    """
    Assert that ``obj`` is an instance of ``cls``.

    Args:
        obj: an instance
        cls: a class ``obj`` shall be an instance of
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``obj`` is not an instance of ``cls``
    """
    if not isinstance(obj, cls):
        raise AssertionError(
            _format_message(
                message, "{} is not an instance of {}".format(obj, cls)
            )
        )


def assert_not_is_instance(obj, cls, message=None):
    """
    Assert that ``obj`` is not an instance of ``cls``.

    Args:
        obj: an instance
        cls: a class ``obj`` shall not be an instance of
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``obj`` is an instance of ``cls``
    """
    if isinstance(obj, cls):
        raise AssertionError(
            _format_message(
                message, "{} is an instance of {}".format(obj, cls)
            )
        )


assert_is_not_instance = assert_not_is_instance


def assert_regex(text, expected_regex, message=None):
    """
    Assert that ``text`` matches regular expression ``expected_regex``.

    Args:
        text: a text to be matched
        expected_regex: a regular expression to be used
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``text`` does not match regular expression ``expected_regex``
    """
    if not isinstance(expected_regex, (type(""), type(b""))):
        raise TypeError("expected_regex must be a string!")

    assert expected_regex, "expected_regex must not be empty."
    expected_regex = re.compile(expected_regex)

    if not expected_regex.search(text):
        raise AssertionError(
            _format_message(
                message,
                "Regular expression {!r} does not match {!r}".format(
                    expected_regex.pattern,
                    text
                )
            )
        )


def assert_not_regex(text, unexpected_regex, message=None):
    """
    Assert that ``text`` does not match regular expression ``unexpected_regex``.

    Args:
        text: a text to be matched
        unexpected_regex: a regular expression to be used
        message=None (str): the message to prepend to the error message

    Raises:
        AssertionError: if ``text`` matches regular expression ``unexpected_regex``
    """
    if not isinstance(unexpected_regex, (type(""), type(b""))):
        raise TypeError("unexpected_regex must be a string!")

    assert unexpected_regex, "unexpected_regex must not be empty."
    unexpected_regex = re.compile(unexpected_regex)

    match = unexpected_regex.search(text)

    if match:
        raise AssertionError(
            _format_message(
                message,
                "Regular expression {!r} matched {!r} in {!r}".format(
                    unexpected_regex.pattern,
                    text[match.start(): match.end()],
                    text
                )
            )
        )
