# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#
# This file was generated by wrappergenerator.py
#

from functools import wraps
import inspect

from htf._decorators import after as _after
from htf._decorators import before as _before
from htf._decorators import enqueue_exception as _enqueue_exception
from htf._decorators import meets_doors_requirements as _meets_doors_requirements
from htf._decorators import periodic as _periodic
from htf._decorators import raises as _raises
from htf._decorators import requirements as _requirements
from htf._decorators import skip as _skip
from htf._decorators import skip_if as _skip_if
from htf._decorators import stacktrace as _stacktrace
from htf._decorators import test as _test
from htf._decorators import timed as _timed


def after(after_function, *after_args, **after_kwargs):
    """
    The `@after` decorator decorates a method so that `after_function`
    is called after the decorated function is called itself.

    Args:
        after_function (callable): the function to be run after
            the decorated function.
        *after_args (tuple): positional arguments to be passed
            to ``afterFunction``.
        **after_kwargs (dict): keyword-arguments to be passed
            to ``afterFunction``.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_after(after_function, *after_args, **after_kwargs)(func))
            return wrapper
        else:
            return _after(after_function, *after_args, **after_kwargs)(func)

    return pure_python_wrapper


def before(before_function, *before_args, **before_kwargs):
    """
    The `@before` decorator decorates a method so that `before_function`
    is called before the decorated function is called itself.

    Args:
        before_function (callable): the function to be run before
            the decorated function.
        *before_args (tuple): positional arguments to be passed
            to ``beforeFunction``.
        **before_kwargs (dict): keyword-arguments to be passed
            to ``beforeFunction``.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_before(before_function, *before_args, **before_kwargs)(func))
            return wrapper
        else:
            return _before(before_function, *before_args, **before_kwargs)(func)

    return pure_python_wrapper


def enqueue_exception(queue):
    """
    Decorate a method to fetch its exception that is enqueued into an
    exception queue.
    The enqueued exception can be handled elsewhere.
    The caught exception can be raised after enqueueing it.

    Just decorate you method with the queue to be used and
    every exception that is raised within the
    decorated method will be put into the queue
    before the exception is raised again.

    Args:
        queue (queue): the queue to enqueue the exception.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_enqueue_exception(queue=queue)(func))
            return wrapper
        else:
            return _enqueue_exception(queue=queue)(func)

    return pure_python_wrapper


enqueueException = enqueue_exception


def meets_doors_requirements(base_url, *requirements):
    """
    The ``@meets_doors_requirements`` decorator reports which requirements
    are checked by a test in the test report
    shown in the :mod:`~htf.report.doors_testreport`
    and the :mod:`~htf.report.html_testreport.HTMLTestReport`.

    Args:
        base_url (str): the DOORS base url the links are created with.
        *requirements (str): a tuple of strings that name the
            checked requirements, eg. "REQ_1", "REQ_2", etc.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_meets_doors_requirements(base_url, *requirements)(func))
            return wrapper
        else:
            return _meets_doors_requirements(base_url, *requirements)(func)

    return pure_python_wrapper


meetsDOORSRequirements = meets_doors_requirements


def periodic(period, maximum_period=None, run_condition_lambda=None, raise_exception=True):
    """
    Decorate a method to be run periodically.

    Args:
        period (float): the period in seconds.
        maximum_period=None (float): if set to a float >= `period`
            the periodic method may take up to `maximum_period` time
            without raising an exception. A warning is printed on
            stdout instead.
        run_condition_lambda=None (callable): a callable (method or
            lambda expression) that has to return true while the method
            is run periodically.
        raise_exception=True (bool): if set to ``True`` an exception is raised
            if the called method takes longer
            than the period (or ``maximum_period`` if set).

    Warning:
        Non-realtime operating systems will not ensure timing accuracy.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(
                _periodic(period=period, maximum_period=maximum_period, run_condition_lambda=run_condition_lambda,
                          raise_exception=raise_exception)(func))
            return wrapper
        else:
            return _periodic(period=period, maximum_period=maximum_period, run_condition_lambda=run_condition_lambda,
                             raise_exception=raise_exception)(func)

    return pure_python_wrapper


def raises(*exceptions):
    """
    Decorate a method to catch different exceptions.

    Args:
        *exceptions (list of exceptions): a tuple of exceptions that are
            fetched.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_raises(*exceptions)(func))
            return wrapper
        else:
            return _raises(*exceptions)(func)

    return pure_python_wrapper


def requirements(*requirements):
    """
    The ``@requirements`` decorator reports which requirements
    are checked by a test in the test report
    shown in the :mod:`~htf.report.html_testreport.HTMLTestReport`.

    Args:
        *requirements (str): a tuple of strings that name the
            checked requirements, eg. "REQ_1", "REQ_2", etc.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_requirements(*requirements)(func))
            return wrapper
        else:
            return _requirements(*requirements)(func)

    return pure_python_wrapper


def skip(reason):
    """
    Decorate a method to be skipped in view of the occasion.

    Args:
        reason (str): the reason to be put into the test report.

    Raises:
        SkipTest: to skip the test.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_skip(reason=reason)(func))
            return wrapper
        else:
            return _skip(reason=reason)(func)

    return pure_python_wrapper


def skip_if(condition, reason):
    """
    Decorate a method to be skipped. The ``condition`` is evaluated at runtime.
    Note that ``condition`` is evaluated twice by the test runner and the test itself. Thus it should not
    introduce any side effects.

    Args:
        condition (bool or callable): the condition may be ``True``,
            ``False`` or a callable object that must return
            ``True`` or ``False``.
            The decorated test is skipped if ``condition`` is
            ``True`` or returns ``True``.
        reason (str): the reason to be put into the test report.

    Raises:
        SkipTest: to skip the test if ``condition`` is or returns ``True``.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_skip_if(condition=condition, reason=reason)(func))
            return wrapper
        else:
            return _skip_if(condition=condition, reason=reason)(func)

    return pure_python_wrapper


skipIf = skip_if


def stacktrace(func):
    """
    The ``@stacktrace`` decorator decorates a method so that its call stack
    is printed to stdout every time it is called.
    This is useful for debugging.
    """
    return _stacktrace(func)


def test(func):
    """
    Mark a function or method as a test.
    """
    return _test(func)


def timed(limit):
    """
    Decorate a method to ensure that it ends within a given time limit.

    Args:
        limit (float): the time limit in seconds.
    """

    def pure_python_wrapper(func):  # pragma: no cover
        if not inspect.isclass(func):
            wrapper = wraps(func)(_timed(limit=limit)(func))
            return wrapper
        else:
            return _timed(limit=limit)(func)

    return pure_python_wrapper
