# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import fnmatch
import inspect
import itertools
import os
import re
import sys
import traceback
from pydoc import locate
from functools import reduce
import glob
import pkgutil

from htf._decorators import _is_cython
from htf.termcolors import system
from htf.gherkin.gherkin_steps import GherkinSteps
from htf.gherkin.feature_parser import GherkinFeatureParser
from htf.core import Test
from htf.core import GherkinScenarioTest, GherkinScenarioOutlineTest


import importlib.util

def load_module(file):
    module_name = os.path.splitext(os.path.basename(file))[0]
    spec = importlib.util.spec_from_file_location(module_name, file)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    try:
        spec.loader.exec_module(module)
    except Exception as exc:  # pragma: no cover
        exc.__cause__ = None
        raise DiscoveryError('Loading module "{}" failed with the following error:\n\n'
                             '{}'.format(file, ''.join(traceback.format_exception(*sys.exc_info()))))
    return module


class DiscoveryError(Exception):
    def __init__(self, message):
        super(DiscoveryError, self).__init__(message)
        self.__cause__ = None  # disable exception chaining in Python 3


class Discovery:
    """
    A class responsible for resolving given specifiers into corresponding test and fixture objects.
    """

    def __init__(self, test_specifiers, fixture_specifiers, gherkin_specifiers, file_pattern='test*.py'):
        """
        Args:
            test_specifiers(list of str): strings specifying tests or test locations
            fixture_specifiers(list of str):  strings specifying fixtures or fixture locations
             gherkin_specifiers(list of str):  strings specifying gherkin statements or their locations
            file_pattern (str): pattern to use to discover files in given folders
        """
        self._test_specifiers = test_specifiers
        self._fixture_specifiers = fixture_specifiers
        self._gherkin_specifiers = gherkin_specifiers
        self._file_pattern = file_pattern

    @staticmethod
    def _get_modules_from_file_specifier(file):
        """
        Return a list of modules given a file specifier. The file specifier can contain wildcards.

        Args:
            file (str): file specifier (can be a wildcard)

        Returns:
            List of module objects
        """
        modules = []
        path = os.path.abspath(file)
        files = glob.glob(path)
        for file in files:
            modules.append(load_module(file))
        return modules

    def _resolve_specifiers(self, specifiers, do_print=True, filter_files=False):
        """
        Resolves specifiers into modules, classes or functions depending on the specifier.

        Args:
            specifiers(list of str): specifiers to resolve

        Returns:
            List of modules, classes and functions
        """
        resolved_specifiers = []
        for specifier in specifiers:
            if do_print is True:
                print(system(specifier))

            if isinstance(specifier, (type(""), bytes)):
                try:
                    if os.path.isdir(specifier):
                        # handle directory specifiers
                        sys.path.append(specifier)
                        for root, dirs, files in os.walk(specifier, topdown=False):
                            for name in files:
                                if name.lower().endswith('.feature'):
                                    resolved_specifiers.append(os.path.join(root, name))
                        resolved_specifiers += self._get_modules_from_directory(specifier, filter_files)
                        continue
                except TypeError as ex:  # pragma: no cover
                    raise
                    pass

                try:
                    # handle objects
                    obj = locate(specifier)
                except Exception as exc:
                    exc.__cause__ = None  # disable exception chaining of the pydoc error
                    raise DiscoveryError('Locating specifier "{}" failed with the following error:\n\n'
                                         '{}'.format(specifier, ''.join(traceback.format_exception(*sys.exc_info()))))
                if obj is not None:
                    resolved_specifiers.append(obj)
                    continue

                try:
                    # handle files and wildcards
                    if os.path.isfile(specifier) or "*" in specifier:
                        if specifier.lower().endswith('.feature'):
                            # assert False, specifier
                            resolved_specifiers += GherkinDiscoveryUtils.get_features_from_file_specifier(specifier)
                            continue
                        resolved_specifiers += self._get_modules_from_file_specifier(specifier)
                        continue
                except TypeError:  # pragma: no cover
                    pass

                raise DiscoveryError('Could not resolve specifier: {}'.format(specifier))

            resolved_specifiers.append(specifier)
        return resolved_specifiers

    def _get_modules_from_directory(self, directory, filter_files):
        """
        Return a list of modules for a given directory.

        Args:
            directory (str): path to the directory

        Returns:
            List of module objects
        """
        modules = []
        path = os.path.abspath(directory)
        files = []
        for root, dirnames, filenames in os.walk(path):
            if filter_files:
                for filename in fnmatch.filter(filenames, self._file_pattern):
                    files.append(os.path.join(root, filename))
            else:
                for filename in fnmatch.filter(filenames, '*.py'):
                    files.append(os.path.join(root, filename))

        for file in files:
            modules.append(load_module(file))
        return modules

    @staticmethod
    def _add_cwd_to_path():
        cwd = os.getcwd()
        if cwd not in sys.path:
            sys.path.insert(0, cwd)

    def discover_gherkin_statements(self):
        """
        Discover gherkin statements

        Returns:
            FunctionDictionaries object
        """
        self._add_cwd_to_path()
        gherkin_steps = GherkinSteps()

        for specifier in self._resolve_specifiers(self._gherkin_specifiers, do_print=False):
            if inspect.ismodule(specifier):
                gherkin_steps.combine(GherkinDiscoveryUtils.get_gherkin_statements_from_module(specifier))

            elif inspect.isclass(specifier) or inspect.isfunction(specifier):
                gherkin_steps.combine(GherkinDiscoveryUtils.get_gherkin_statements_from_object(specifier))

        return gherkin_steps

    def discover_tests(self):
        """
        Discover test functions using the test specifiers provided.

        Returns:
            List of test functions
        """

        self._add_cwd_to_path()
        gherkin_steps = self.discover_gherkin_statements()
        tests = []
        for specifier in self._resolve_specifiers(self._test_specifiers, filter_files=True):
            if inspect.ismodule(specifier):
                tests += TestDiscoveryUtils.get_tests_from_module(specifier, self._file_pattern)

            elif TestDiscoveryUtils.is_test(specifier):
                tests.append(Test(specifier))

            elif isinstance(specifier, str) and os.path.isfile(specifier):  # is feature file
                feature = GherkinFeatureParser(filename=specifier).get_feature_from_file()
                tests.extend(GherkinDiscoveryUtils.create_tests_from_feature(feature, gherkin_steps))
            else:
                tests += TestDiscoveryUtils.get_tests_from_object(specifier)

        return tests

    def discover_fixtures(self):
        """
        Discover fixtures using the fixture specifiers provided.

        Returns:
            List of fixtures
        """
        self._add_cwd_to_path()
        fixtures = []

        for specifier in self._resolve_specifiers(self._fixture_specifiers, do_print=False):
            if inspect.ismodule(specifier):
                fixtures += list(set(FixtureDiscoveryUtils.get_fixtures_from_module(specifier)))

            elif inspect.isclass(specifier) or inspect.isfunction(specifier):
                fixtures += FixtureDiscoveryUtils.get_fixtures_from_object(specifier)

        return fixtures


class TestDiscoveryUtils:
    @staticmethod
    def is_test(obj):
        """
        Checks if a given object should be considered a test.

        Args:
            obj (object): object to inspect

        Returns:
            True if the object is likely a test, False otherwise
        """
        return (inspect.isfunction(obj) or inspect.ismethod(obj) or _is_cython(obj)) \
               and (obj.__name__.startswith('test') or hasattr(obj, "htf_test")) \
               and not hasattr(obj, "htf_fixture")

    @staticmethod
    def get_tests_from_object(obj):
        """
        Returns a list of test functions for a given object. A test function is any function or method whose name
        starts with 'test'.

        Args:
            obj (object): object to look for tests in

        Returns:
            List of test functions
        """

        return [Test(test, obj) for test in [getattr(obj, attr) for attr in dir(obj)] if
                TestDiscoveryUtils.is_test(test)]

    @staticmethod
    def get_tests_from_module(module, file_pattern):
        """
        Return a list of test functions for a given module.

        Args:
            module: module to look for tests in
            file_pattern: pattern to use to discover files in given folders

        Returns:
            List of test functions
        """
        tests = []
        if hasattr(module, '__path__'):
            for mod in pkgutil.walk_packages(module.__path__, prefix=module.__name__ + '.'):
                module_pattern = file_pattern.replace('.py', '').replace('*', '.*')
                if re.match(module_pattern, mod[1].split('.')[-1]):
                    try:
                        print("--------------- import module", mod[1], "from", mod)
                        imported = __import__(mod[1], fromlist='dummy')
                    except Exception as exc:  # pragma: no cover
                        exc.__cause__ = None
                        raise DiscoveryError('Importing test module "{}" failed with the following error:\n\n'
                                             '{}'.format(module, ''.join(traceback.format_exception(*sys.exc_info()))))
                    tests += TestDiscoveryUtils.get_tests_from_module(imported, file_pattern)
        tests += list(itertools.chain(*map(TestDiscoveryUtils.get_tests_from_object,
                                           [val for name, val in
                                            inspect.getmembers(module, predicate=inspect.isclass)])))
        tests += TestDiscoveryUtils.get_tests_from_object(module)
        return tests


class FixtureDiscoveryUtils:
    @staticmethod
    def is_fixture(obj):
        return (inspect.isfunction(obj) or inspect.isclass(obj)) \
               and hasattr(obj, 'htf_fixture') \
               and not hasattr(obj, "htf_test")

    @staticmethod
    def get_fixtures_from_object(obj):
        """
        Return a list of fixtures for a given object. A fixture is any function or class
        with the 'htf_fixture' attribute.

        Args:
            obj: object to look for fixtures in

        Returns:
            List of fixtures
        """

        return [obj] if FixtureDiscoveryUtils.is_fixture(obj) else []

    @staticmethod
    def get_fixtures_from_module(module):
        """
        Return a list of fixtures for a given module.

        Args:
            module: module to look for fixtures in

        Returns:
            List of fixtures
        """
        fixtures = []
        if hasattr(module, '__path__'):
            for mod in pkgutil.walk_packages(module.__path__, prefix=module.__name__ + '.'):
                try:
                    imported = __import__(mod[1], fromlist=str('dummy'))
                except Exception as exc:  # pragma: no cover
                    exc.__cause__ = None
                    raise DiscoveryError('Importing fixture module "{}" failed with the following error:\n\n'
                                         '{}'.format(module, ''.join(traceback.format_exception(*sys.exc_info()))))
                fixtures += FixtureDiscoveryUtils.get_fixtures_from_module(imported)
        fixtures += list(itertools.chain(*map(FixtureDiscoveryUtils.get_fixtures_from_object,
                                              [val for name, val in
                                               inspect.getmembers(module, predicate=inspect.isclass)])))
        fixtures += list(itertools.chain(*map(FixtureDiscoveryUtils.get_fixtures_from_object,
                                              [val for name, val in
                                               inspect.getmembers(module, predicate=inspect.isfunction)])))
        return fixtures


class GherkinDiscoveryUtils:
    @staticmethod
    def get_gherkin_statements_from_object(obj):
        """
          Returns a dictionary of gherkin statements that capture the function.

          Args:
              obj: object to look for statements in

          Returns:
              a dictionary of gherkin statements that capture the function
        """
        gherkin_steps = GherkinSteps()

        # check for feature
        if hasattr(obj, "_htf_feature"):
            gherkin_steps.register_feature(obj)

        # check for scenario
        if hasattr(obj, "_htf_scenario"):
            gherkin_steps.register_scenario(obj)

        # check for given
        if hasattr(obj, "_htf_given"):
            gherkin_steps.register_given_precondition(obj)

        # check for when
        if hasattr(obj, "_htf_when"):
            gherkin_steps.register_when_condition(obj)

        # check for then
        if hasattr(obj, "_htf_then"):
            gherkin_steps.register_then_assertion(obj)

        # check for transformation
        if hasattr(obj, "_htf_transformation"):
            gherkin_steps.register_transformation(obj)

        # check for data generator
        if hasattr(obj, "_htf_data_generators"):
            gherkin_steps.register_data_generator(obj)

        return gherkin_steps

    @staticmethod
    def get_gherkin_statements_from_module(module):
        """
        Returns a dictionary of gherkin statements capture the gherkin functions the module.

        Args:
            module: module to search inside

        Returns:
            Dictionary of gherkin statements
        """
        function_dictionaries = GherkinSteps()
        if hasattr(module, '__path__'):
            for mod in pkgutil.walk_packages(module.__path__, prefix=module.__name__ + '.'):
                try:
                    imported = __import__(mod[1], fromlist=str('dummy'))
                except Exception as exc:  # pragma: no cover
                    exc.__cause__ = None
                    raise DiscoveryError('Importing fixture module "{}" failed with the following error:\n\n'
                                         '{}'.format(module, ''.join(traceback.format_exception(*sys.exc_info()))))

                function_dictionaries.combine(GherkinDiscoveryUtils.get_gherkin_statements_from_module(imported))

        def dict_combiner(a,b):
            assert(isinstance(a, GherkinSteps))
            assert (isinstance(b, GherkinSteps))
            a.combine(b)
            return a

        return reduce(dict_combiner,
                      [o for o in map(
                          GherkinDiscoveryUtils.get_gherkin_statements_from_object,
                          [val for name, val in inspect.getmembers(module, predicate=inspect.isfunction)]
                      )], function_dictionaries)

    @staticmethod
    def get_features_from_file_specifier(pattern):
        """
        Return a list of features filenames. The file specifier can contain wildcards.

        Args:
            pattern (str): file specifier (can be a wildcard)

        Returns:
            List of filenames
        """
        modules = []
        files = glob.glob(pattern)
        return list(files)

    @staticmethod
    def create_tests_from_feature(feature, gherkin_steps):
        return [
                   GherkinDiscoveryUtils.create_test_from_scenario(feature, scenario, gherkin_steps)
                   for scenario in feature.scenarios
               ] + \
               [
                   GherkinDiscoveryUtils.create_data_driven_test_from_scenario_outline(feature, scenario_outline, gherkin_steps)
                   for scenario_outline in feature.scenario_outlines
               ]

    @staticmethod
    def create_test_from_scenario(feature, scenario, gherkin_steps):
        return GherkinScenarioTest(feature, scenario, gherkin_steps)

    @staticmethod
    def create_data_driven_test_from_scenario_outline(feature, scenario_outline, gherkin_steps):
        return GherkinScenarioOutlineTest(feature, scenario_outline, gherkin_steps)

