# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import inspect
import sys
import types
from collections import OrderedDict

from htf.event import EventManager

_current_environment = None  # access the current environment


def get_current_environment():
    global _current_environment
    return _current_environment


class TestEnvironment:
    """
    A TestEnvironment manages test fixtures and their life-times depending on received events.
    """

    def __init__(self):
        """
        """
        self._fixtures = OrderedDict()  # fixtures are generator functions and yielding a fixture object
        self._fixture_objects = {}  # objects yielded by fixture generators
        self._generator_fixtures = {}  # generator objects instantiated by calling fixtures
        self._function_fixtures = {}  # return values of function fixtures
        self._active_scopes = set()  # scopes for which we received '*_started' events, but not '*_finished'
        self._fixture_instantiation_orders = {
            'test': [],
            'class': [],
            'module': [],
            'session': []
        }  # a dictionary which holds the orders of fixture instantiation for every possible scope to be able to teardown dependant fixtures in the reversed order

        global _current_environment
        _current_environment = self

    def get_fixture(self, name, requesting_scope='test', resolving=None):
        """
        Returns a fixture object as identified by the name parameter. If a fixture requires other fixtures
        they, too, will be instantiated if necessary.

        Args:
            name (str): name of the requested fixture
            requesting_scope='test' (str): the scope of the requesting fixture
            resolving(None or list): an internal parameter used to track the fixture names being actively resolved

        Returns:
            fixture instance

        Raises:
            Exception if a circular dependency is detected during resolution
        """
        scopes = ['test', 'class', 'module', 'session']
        assert requesting_scope in scopes, 'Possible values for scope: session, module, class, test'

        if resolving is None:
            resolving = []
        if name in resolving:
            raise Exception('Circular dependency in requested fixture: {}.'.format(name))
        else:
            resolving.append(name)

        # check if requesting scope is less or equal the fixtures' scope
        try:
            if scopes.index(requesting_scope) > scopes.index(self._fixtures[name]._fixture_scope):
                raise Exception("Fixture '{}' with scope '{}' cannot be requested in a wider scope '{}'.".format(
                    name,
                    self._fixtures[name]._fixture_scope,
                    requesting_scope
                ))
        except KeyError:
            raise Exception('Could not find fixture: {}.'.format(name))

        if name not in self._fixture_objects:
            self._instantiate_fixture(name, resolving)

        resolving.remove(name)

        return self._fixture_objects[name]

    def set_fixture(self, name, fixture):
        """
        Registers a fixture with a name it can be requested by.

        Args:
            name: name of the fixture to register
            fixture: fixture generator

        Returns:

        """
        self._fixtures[name] = fixture
        EventManager.register_callback(fixture._fixture_scope + '_started', self.handle_event)
        EventManager.register_callback(fixture._fixture_scope + '_finished', self.handle_event)

    def _instantiate_fixture(self, name, resolving):
        assert name not in self._fixture_objects and name not in self._generator_fixtures, \
            'The requested fixture ({}) was already instantiated.'.format(name)

        scope = self._fixtures[name]._fixture_scope

        if scope in self._active_scopes:
            args = self._get_args(self._fixtures[name])
            if len(args) > 0:
                fixture = self._fixtures[name](*[self.get_fixture(n, self._fixtures[name]._fixture_scope, resolving) for n in args])
            else:
                fixture = self._fixtures[name]()

            if isinstance(fixture, types.GeneratorType) or callable(getattr(fixture, "__next__", None)):
                # is a generator or mimics the generator protocol
                self._fixture_objects[name] = next(fixture)
                self._generator_fixtures[name] = fixture
                self._fixture_instantiation_orders[scope].append(name)
            else:
                # fixture is not a generator
                self._fixture_objects[name] = fixture
                self._function_fixtures[name] = fixture
                self._fixture_instantiation_orders[scope].append(name)
        else:
            raise Exception('Fixture {} was requested outside of its scope of {}.'.format(name, self._fixtures[name]._fixture_scope))

    @staticmethod
    def _get_args(obj):
        """
        Returns a list of arguments of an object, fixtures in this case. Used to determine if a fixture is
        dependent on another fixture.

        Args:
            obj: object to get arguments for

        Returns:
            List of str of arguments
        """
        args = inspect.getfullargspec(obj).args

        if inspect.isclass(obj):
            args = args[1:]

        return args

    def _clean_up_generator_fixture(self, name):
        """
        Remove any generator fixture objects and active generator objects produced by a fixture.

        Args:
            name: name of the fixture to clean up
        """
        del self._generator_fixtures[name]
        del self._fixture_objects[name]

    def _clean_up_function_fixture(self, name):
        """
        Remove any function fixture objects

        Args:
            name: name of the fixture to clean up
        """
        del self._function_fixtures[name]
        del self._fixture_objects[name]

    def _process_state_finished(self, scope):
        """
        When a '*_finished' event is received all fixtures matching the scope are cleaned up if necessary.
        If an exception is raised during the clean up of a fixture it is cached while any other remaining
        fixtures are cleaned up and re-raised after. If multiple exceptions are raised during clean up
        only the exception which occurred first will be re-raised.

        Raises:
            re-raises any exceptions raised during clean up of fixtures
        """
        exception = None

        for name in reversed(self._fixture_instantiation_orders[scope]):
            if self._fixtures[name]._fixture_scope not in self._active_scopes:  # pragma: no branch
                if name in self._generator_fixtures:
                    try:
                        next(self._generator_fixtures[name])
                        raise Exception('Fixture ({}) did not raise StopIteration on tear down.'.format(name))
                    except StopIteration:
                        self._clean_up_generator_fixture(name)
                    except Exception as ex:
                        self._clean_up_generator_fixture(name)
                        ex._htf_exc_info = sys.exc_info()  # Python 2.7 doesn't cache exception tracebacks
                        exception = exception or ex
                if name in self._function_fixtures:
                    self._clean_up_function_fixture(name)

        # clear instantiation order for the scope
        self._fixture_instantiation_orders[scope] = []

        if exception is not None:
            raise exception

    def _process_state_started(self, scope):
        """
        When a '*_started' event is received all fixtures which have the 'auto' attribute set to True are instantiated.
        """
        for name, fixture in list(self._fixtures.items()):
            if fixture._fixture_scope == scope and fixture._fixture_auto is True:
                self.get_fixture(name, requesting_scope=scope)

    def _process_state_change(self, scope, state):
        """
        Handles scope state changes.
        When a 'started' state is received the scope is added to a set of active scopes.
        On 'finished' states the scope is removed from the set of active scopes and clean up
        is started for any fixtures matching the scope that was left.

        Args:
            scope (str): the scope which changed state
            state (str): 'started' or 'finished'

        Raises:
            Exception if a 'finished' state is received for a non-active scope.
        """
        if state == 'started':
            self._active_scopes.add(scope)
            self._process_state_started(scope)
        elif state == 'finished':  # pragma: no branch
            try:
                self._active_scopes.remove(scope)
            except KeyError:
                raise Exception("Scope '{}' is not active.".format(scope))
            self._process_state_finished(scope)

    def handle_event(self, event):
        """
        Splits events like 'session_started' or 'test_finished' into the scope and state and processes
        the state change.

        Args:
            event (str): the event that occurred
        """
        scope, state = event.type.split('_')
        self._process_state_change(scope, state)

    def get_active_scopes(self):
        """
        Returns a set of scopes which are currently active

        Returns:
            set of strings representing active scopes
        """
        return self._active_scopes