# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import datetime
import sys
import traceback
from htf.exceptions import SkipTest, NotRunException, ScheduledException
from htf.termcolors import failure, error, success, skipped


class Evaluation:
    """
    A context-manager to interpret the result of a test executed in its context.
    """
    def __init__(self, result_type):
        """
        Initialize the instance by setting the result type to generate on call.

        Args:
            result_type (class): the type of result to generate
        """
        self._result_type = result_type
        self.result = None
        self.silent = False

    def __enter__(self):
        """
        Returns:
            instance of the result type
        """
        return self.result

    def __call__(self, test):
        """
        On call the result type is passed the test it belongs to, and result is generated
        Also enables capture of any output generated with the evaluation context.

        Args:
            test (function): test function the result belongs to

        Returns:
            the Evaluation instance

        """
        self.result = self._result_type(test)
        self.result.set_output_capture_enable(enabled=True)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        On leaving the context block the result is interpreted by examining the type of exception, if any,
        was raised within the context.

        Args:
            exc_type: exception type
            exc_val: exception value
            exc_tb: exception traceback

        Returns:
            True to mark any exceptions as handled
        """
        self.result.set_exception(exc_type, exc_val, exc_tb)

        message = ''

        if exc_type is None:
            self.result.result = 'success'
            message += success('SUCCESS')

        elif exc_type == AssertionError:
            self.result.result = 'failure'
            message += failure('FAILURE') + '\n'
            message += failure(''.join(traceback.format_exception(self.result.exception_type, self.result.exception_value,
                                       self.result.exception_traceback)))

        elif issubclass(exc_type, SkipTest):
            self.result.result = 'skipped'
            message += skipped('SKIPPED') + '\n'
            message += skipped(''.join(traceback.format_exception(self.result.exception_type, self.result.exception_value,
                                       self.result.exception_traceback)))

        elif issubclass(exc_type, NotRunException):
            self.result.result = 'skipped'

        elif issubclass(exc_type, ScheduledException):
            self.result.result = 'scheduled'

        else:
            self.result.result = 'error'
            message += error('ERROR') + '\n'
            message += error(''.join(traceback.format_exception(self.result.exception_type, self.result.exception_value,
                                     self.result.exception_traceback)))

        if not self.silent and message:
            try:
                print(message)
            except UnicodeEncodeError:  # pragma: no cover
                print(message.encode(encoding=sys.getdefaultencoding(), errors='replace').
                      decode(encoding=sys.getdefaultencoding()))

        self.result.stop_time = datetime.datetime.now()

        self.result.set_output_capture_enable(enabled=False)

        return True
