# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

class Event:
    """
    An event object with a type and optional additional data.
    """
    def __init__(self, event_type, kwargs=None):
        """
        Args:
            event_type (str): type of the event
            kwargs (dict): optional additional data related to the event.
        """
        self.type = event_type
        self.kwargs = kwargs or {}

    def __str__(self):
        return "{}: {}".format(self.type, self.kwargs)


class EventManager:
    """
    A simple EventManager handling callbacks and events.
    """
    def __init__(self):
        self._callbacks = {}

    def register_callback(self, event_type, callback):
        """
        Register a callback function for an event type.

        Args:
            event_type (str): event type to register the callback for
            callback (callable): function to call if a matching event occurs
        """
        if event_type not in self._callbacks:
            self._callbacks[event_type] = []
        if callback not in self._callbacks[event_type]:
            self._callbacks[event_type].append(callback)

    def unregister_callback(self, event_type, callback):
        """
        Remove a callback for an event type.

        Args:
            event_type (str): event type to unregister the callback for
            callback (callable): function to remove
        """
        try:
            self._callbacks[event_type].remove(callback)
        except ValueError:
            pass

    def handle_event(self, event):
        """
        Calls all registered callback functions matching the event until a callback returns True,
        marking the event as handled.

        Args:
            event (Event): event that occurred
        """
        # print(event)
        if event.type in self._callbacks:
            for callback in self._callbacks[event.type]:
                if callback(event):
                    break

    def clear_callbacks(self):
        """
        Clear all callbacks.
        """
        self._callbacks = {}


EventManager = EventManager()
