# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#
import collections
import json
import logging
import base64
import multiprocessing
import sys
import threading
from collections import namedtuple

from htf.core import ThreadsController
import htf.assertions
from htf.core import fixture
import time
import jira as jira_module


class _ClassFixture:
    def __iter__(self):  # pragma: no cover
        self._raise = False
        return self

    def _cleanup(self):
        pass

    def get_fixture(self):
        return self

    def __next__(self):
        if not hasattr(self, '_raise'):
            self._raise = False
        if self._raise is False:
            self._raise = True
            return self.get_fixture()
        else:
            self._cleanup()
            raise StopIteration()


@fixture('session')
def assertions():
    """
    A fixture to access :module:`htf.assertions`.
    """
    return htf.assertions


@fixture('test')
def context():
    """
    A context fixture used to share context data in behavior driven steps.
    """
    class Context:
        pass

    return Context()


@fixture('test')
class threads(_ClassFixture):
    """
    Fixture for threads with scope 'test' that are stopped automatically and can run in background and periodically.
    """

    def __init__(self):
        self._threads_controller = ThreadsController()

    def _cleanup(self):
        self.stop()
        self._threads_controller.check_for_exceptions()

    def run_background(self, func, name=None, stop=None, force_stop=True, join_timeout=1, args=None,
                       kwargs=None):
        """
        With this method you can run another method in background.

        Args:
            func (callable): the method to be run in background.
            name=None (str): a name for the background task.
                Default: ``None``. If `name` is ``None`` `func.__name__` is
                used.
            stop=None (callable): callable to stop the thread
            force_stop=True (bool): force the thread to stop when leaving its scope (can impact performance)
            join_timeout=1 (float): the timeout to be used when calling ``join``
                on the thread.
            args=None (tuple): a tuple of positional arguments that are passed
                to the background task when it is called.
            kwargs=None (dict): a dictionary of keyword arguments that are
                passed to the background task when it is called.

        Returns:
            stop-trigger (callable): a callable object is returned that stops
                execution when called.
        """

        return self._threads_controller.run_background(func=func, name=name, stop=stop, force_stop=force_stop,
                                                       join_timeout=join_timeout, args=args, kwargs=kwargs)

    def run_periodic(self,
                     func,
                     period,
                     maximum_period=None,
                     name=None,
                     raise_exception=True,
                     args=None,
                     kwargs=None):
        """
        With this method you can run another method periodically
        in background with a given period.

        Args:
            func (callable): the method to be called periodically in
                background.
            period (float): the period in second.
            maximum_period=None (float): if set to a float >= ``period``
                ``func`` may take up to ``maximum_period`` time without
                raising an exception. A warning is printed out to stdout
                instead.
            name=None (str): a name for the background task.
            raise_exceptions=True: if set to ``True`` exceptions are put into
                the exception queue to let a test fail if a background thread
                fails.
            args=None (tuple): a tuple of positional arguments that are passed
                to the background task when it is called.
            kwargs=None (dict): a dictionary of keyword arguments that are
                passed to the background task when it is called.

        Returns:
            stop-trigger (callable): a callable object is returned that stops
                execution when called.
        """

        return self._threads_controller.run_periodic(
                     func=func,
                     period=period,
                     maximum_period=maximum_period,
                     name=name,
                     raise_exception=raise_exception,
                     args=args,
                     kwargs=kwargs)

    def stop_thread(self, name_or_function):
        """
        This method stops a specific thread with the name or
        function described by ``name_or_function`` by calling
        :func:`~htf.Thread.join` with join timeout supplied
        by :func:`~htf.fixtures.threads.run_periodic` or
        :func:`~htf.fixtures.threads.run_background` or on it.
        It does not set a stop condition so the thread must end and must
        not include an endless loop.
        If the thread is still alive after the first join messages are
        printed to inform the user.

        Args:
            name_or_function (str, callable): the name or the function
                (reference) of the thread to be stopped.

        Returns:
            int: the number of stopped threads.

        Raises:
            Exception: if the thread was not found
        """
        return self._threads_controller.stop_thread(name_or_function=name_or_function)

    def stop(self):
        """
        This method stops all background threads by calling ``thread.join()``
        until it is not alive anymore.
        """
        return self._threads_controller.stop()


@fixture('session')
class session_threads(threads):
    """
    Fixture for threads with scope 'session' that are stopped automatically and can run in background and periodically.
    """


class ProcessesController:
    """
    Processes controller to start and stop background processes.
    """

    def __init__(self):
        self.lock = threading.Lock()
        self.processes = {}

    def run(self, group=None, target=None, name=None, args=(), kwargs=None, daemon=None,
              timeout=1.0):
        """
        Run a background process.

        Args:
            group=None: the group
            target=None (callable): A callable to be run in background. Must be pickleable.
            name=None (str): the name of the background process.
            args=() (tuple): the positional arguments passed to the background process callable.
            kwargs=None (dict): the keyword arguments passed to the background process callable.
            daemon=None (bool): if set to ``True`` the child process is terminated if the
                parent process ends.
            timeout=1.0 (float): the timeout for joining, terminating and killing the process.

        Returns:
            the process
        """
        if kwargs is None:
            kwargs = {}

        process = multiprocessing.Process(
            group=group, target=target, name=name, args=args, kwargs=kwargs, daemon=daemon)

        with self.lock:
            self.processes[process] = {
                "type": "process",
                "group": group,
                "name": name,
                "targat": target,
                "args": args,
                "kwargs": kwargs,
                "daemon": daemon,
                "timeout": timeout
            }

        process.start()

        return process

    def pool(self, processes=None, initializer=None, initargs=(), maxtasksperchild=None, timeout=1.0):
        """
        Create a pool to run tasks in parallel.

        Args:
            processes=None (int): the number of child processes to be started.
            initializer=None (callable): f initializer is not None then each worker process will call
                ``initializer(*initargs)`` when it starts.
            initargs=() (tuple): arguments for ``initializer``.
            maxtasksperchild=None (int): the maximum number of tasks per child process.
            timeout=1.0 (float): the timeout for joining the pool process.

        Returns:
            the pool
        """
        pool = multiprocessing.Pool(
            processes=processes, initializer=initializer, initargs=initargs,
            maxtasksperchild=maxtasksperchild
        )

        with self.lock:
            self.processes[pool] = {
                "type": "pool",
                "timeout": timeout
            }

        return pool

    def stop(self):
        with self.lock:
            for process, data in self.processes.items():

                # try to join the process
                timeout = data["timeout"]
                if timeout is not None:  # pragma: no branch
                    if data["type"] == "process":
                        process.join(timeout=timeout)

                        if process.is_alive():
                            # terminate process
                            process.terminate()
                            time.sleep(timeout)

                        if process.is_alive():  # pragma: no cover
                            # kill process
                            process.kill()
                            time.sleep(timeout)

                        assert not process.is_alive(), "The process could not be stopped!"
                    else:
                        process.close()
                        time.sleep(timeout)
                        process.join()


@fixture('test')
class processes(_ClassFixture):
    """
    Fixture for background processes with scope ``'test'`` that are stopped automatically.
    Uses :mod:`multiprocessing`.
    """
    def __init__(self):
        self._processes_controller = ProcessesController()

    def _cleanup(self):
        self._processes_controller.stop()

    def run(self, group=None, target=None, name=None, args=(), kwargs=None, daemon=None,
            timeout=1.0):
        """
        Run a background process.

        Args:
            group=None: the group
            target=None (callable): A callable to be run in background. Must be pickleable.
            name=None (str): the name of the background process.
            args=() (tuple): the positional arguments passed to the background process callable.
            kwargs=None (dict): the keyword arguments passed to the background process callable.
            daemon=None (bool): if set to ``True`` the child process is terminated if the
                parent process ends.
            timeout=1.0 (float): the timeout for joining, terminating and killing the process.

        Returns:
            the process
        """
        return self._processes_controller.run(
            group=group, target=target, name=name, args=args, kwargs=kwargs, daemon=daemon,
            timeout=timeout)

    def pool(self, processes=None, initializer=None, initargs=(), maxtasksperchild=None, timeout=1.0):
        """
        Create a pool to run tasks in parallel.

        Args:
            processes=None (int): the number of child processes to be started.
            initializer=None (callable): f initializer is not None then each worker process will call
                ``initializer(*initargs)`` when it starts.
            initargs=() (tuple): arguments for ``initializer``.
            maxtasksperchild=None (int): the maximum number of tasks per child process.
            timeout=1.0 (float): the timeout for joining the pool process.

        Returns:
            the pool
        """
        return self._processes_controller.pool(
            processes=processes, initializer=initializer, initargs=initargs,
            maxtasksperchild=maxtasksperchild, timeout=timeout
        )


@fixture('session')
class session_processes(processes):
    """
    Fixture for background processes with scope ``'session'`` that are stopped automatically.
    Uses :mod:`multiprocessing`.
    """

@fixture('test')
class step(_ClassFixture):
    """
    Run a test step that is documented in the test report.
    Scope is ``'test'``.

    Args:
        title (str): the step's title.
        callable_obj=None (callable): a callable object that is called
            within the step or ``None``.
        *args: a tuple of positional arguments for the step's
            context.
        **kwargs: a dictionary of keyword arguments for the step's
            context.
    """
    def __init__(self, result):
        self._result = result

    def __call__(self, title, *args, **kwargs):
        return self._result.step(title, *args, **kwargs)

    def skip(self, reason=None):
        """
        Skip a step.

        Args:
            reason=None (str): the reason for the skipped step

        Raises:
            htf.SkipStep
        """
        raise htf.SkipStep(reason)


@fixture('test')
class attachments(_ClassFixture):
    """
    Attachments to the test report.
    Scope is ``'test'``.
    """
    def __init__(self, result):
        self.result = result

    def set_maximum_attachment_size(self, size):
        """
        Set the maximum attachment size to ``size`` bytes.

        Args:
            size (int): the maximum attachment size in bytes. Default ist 10 MB.
        """
        self.result.set_maximum_attachment_size(size)

    def attach_file(self, filename, title):
        """
        Attach a file to the current test that is put into
        the test report.

        Args:
            filename (str): the filename to be attached
            title (str): the title
        """
        self.result.attach_file(filename, title)

    def attach_url(self, url, title):
        """
        Attach an url to the current test that is put into the test report.
        This results in an external link and should be used for huge files.

        Args:
            url (str): the url to be attached
            title (str): the title
        """
        self.result.attach_url(url, title)

    def attach_bytes(self, title, data_bytes, filename, mime_type):
        """
        Accepts bytes and mime type and converts them to a data url and attaches it to the result.

        Args:
            title (str): the title
            data_bytes (bytes): the bytes to attach
            filename (str): the file name
            mime_type (str): MIME type of the data bytes
        """
        self.result.attach_bytes(title, data_bytes, filename, mime_type)


@fixture('test')
class metadata(_ClassFixture):
    """
    Set metadata in test report.
    Scope is ``'test'``.
    """
    def __init__(self, result):
        self.result = result

    def set(self, name, value):
        """
        Set a metadata name to a value.

        Args:
            name: metadata name
            value: metadata value
        """
        self.result.set_metadata(name, value)

    def get(self, name):
        """
        Get metadata for name.

        Args:
            name: metadata name

        Returns:
            object: the value
        """
        return self.result.get_metadata(name)

    def __setitem__(self, key, value):
        self.result.set_metadata(key, value)

    def __getitem__(self, key):
        return self.result.get_metadata(key)


class Parameters(collections.abc.Mapping):
    """
    Access parameters in test report.
    Scope is ``'session'``. Is instantiated automatically.
    """
    def __init__(self, data):
        self._data = data or dict()

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __str__(self):  # pragma: no cover
        return str(self._data)

    def __repr__(self):  # pragma: no cover
        return repr(self._data)


@fixture('session')
def settings():
    """
    A fixture to access :class:`htf.Settings`.
    """
    from htf import Settings
    yield Settings()


@fixture('session')
class delay(_ClassFixture):
    """
    Delay execution for a given duration and print progress.
    Scope is ``'session'``.

    Args:
        duration (float): delay duration in seconds
    """

    @staticmethod
    def __call__(duration):
        assert duration > 0.0, "duration must be positive and > 0.0"
        counter = int(duration)
        t_0 = time.time()
        while True:
            print('\rTime left: {:5} s ({} s total)'.format(counter, duration), end='')
            counter -= 1
            time_left = max(0, (t_0 + duration) - time.time())
            if time_left < 1:
                break
            time.sleep(1)
        time.sleep(time_left)
        print("")


@fixture(scope='test')
class interaction:  # pragma: no cover
    def __init__(self, interaction_server, result):
        self._interaction_server = interaction_server
        self._result = result

    def _expand_choices(self, choices, enable_checked=False):
        """

        Args:
            choices (list of str or tuples): the choices to be expanded
            enable_checked=False (bool): if set to ``True`` a boolean value can be supplied

        Returns:
            a list of expanded choises to be sent to the front end
        """

        expanded_choices = []

        for choice in choices:
            expanded_choice = None
            if isinstance(choice, (tuple, list)):
                assert 2 <= len(choice) <= 3, "each supplied tuple or list must contain two or three elements"
                if len(choice) == 2:
                    # check if second argument is a bool or a string
                    if choice[1] in [True, False]:
                        # duplicate first element
                        assert enable_checked, "checked values are not allowed"
                        expanded_choice = (choice[0], choice[0], choice[1])
                    else:
                        # set checked to False
                        assert isinstance(choice[0], (type(""), type(b""))), "the supplied choice must be a string"
                        assert isinstance(choice[1], (type(""), type(b""))), "the supplied answer must be a string"
                        expanded_choice = (choice[0], choice[1], False)
                else:
                    # three elements
                    assert enable_checked, "checked values are not allowed"
                    assert isinstance(choice[0], (type(""), type(b"")))
                    assert isinstance(choice[1], (type(""), type(b"")))
                    assert isinstance(choice[2], bool)
                    expanded_choice = choice
            else:
                assert isinstance(choice, (type(""), type(b""))), "the supplied choice must be a string"
                expanded_choice = (choice, choice, False)

            # check types
            assert isinstance(expanded_choice[0], (type(""), type(b"")))
            assert isinstance(expanded_choice[1], (type(""), type(b"")))
            assert isinstance(expanded_choice[2], bool)

            expanded_choices.append(expanded_choice)

        return expanded_choices

    def message_dialog(self, title, text, timeout=300):
        """
        Create an interactive message dialog to display a message.

        Args:
            title (str): the title of the message dialog
            text (str): the content of the message dialog:
            timeout=300 (float): the timeout until the message dialog disappears in seconds (``None`` means no timeout)

        Returns:
            ``"ok"`` in case of a click on the ok button or ``None`` in case of a timeout.
        """
        ret = self._interaction_server.message_dialog(title=title, text=text, timeout=timeout)
        step = {
            "title": title,
            "text": self._interaction_server.text_to_html(text),
            "type": "message_dialog",
            "reason": None,
            "result": ret,
            "steps": []
        }
        self._result._add_step(step)
        return ret

    def waiting_dialog(self, title, text, timeout):
        """
        Create a waiting dialog.

        Args:
            title (str): the title of the waiting dialog
            text (str): the content of the waiting dialog
            timeout (float): the timeout until the waiting dialog disappears in seconds

        Returns:
            ``'Ok'`` after the timeout is over.
        """
        assert timeout is not None, "timeout must not be None"
        ret = self._interaction_server.waiting_dialog(title=title, text=text, timeout=timeout)
        step = {
            "title": title,
            "text": self._interaction_server.text_to_html(text),
            "type": "waiting_dialog",
            "reason": None,
            "result": ret,
            "steps": []
        }
        self._result._add_step(step)
        return ret

    def input_dialog(self, title, text, timeout=300, validator=None):
        """
        Create an input dialog to let the user interact with some textual input.

        Args:
            title (str): the title of the input dialog
            text (str): the content of the input dialog
            timeout=300 (float): the timeout until the input dialog disappears in seconds (``None`` means no timeout)
            validator=None (callable): the validator to validate the inputs on server side

        Returns:
            object or None: the string the user supplied, the return value of the validator or ``None``
                in case of a timeout
        """
        ret = self._interaction_server.input_dialog(title=title, text=text, timeout=timeout, validator=validator)
        step = {
            "title": title,
            "text": self._interaction_server.text_to_html(text),
            "type": "input_dialog",
            "reason": None,
            "result": ret,
            "steps": []
        }
        self._result._add_step(step)
        return ret

    def yes_no_dialog(self, title, text, timeout=300):
        """
        Create a yes-no dialog to ask the user a question.

        Args:
            title (str): the title of the yes-no dialog
            text (str): the content of the yes-no dialog
            timeout=300 (float): the timeout until the yes_no_dialog disappears in seconds (``None`` means no timeout)

        Returns:
            str or None: ``'yes'`` or ``'no'`` depending on the user input or ``None`` in case of a timeout
        """
        ret = self._interaction_server.yes_no_dialog(title=title, text=text, timeout=timeout)
        step = {
            "title": title,
            "text": self._interaction_server.text_to_html(text),
            "type": "yes_no_dialog",
            "reason": None,
            "result": ret,
            "steps": []
        }
        self._result._add_step(step)
        return ret

    def button_dialog(self, title, text, choices, timeout=300):
        """
        Create a dialog with buttons and definable answers.

        Args:
            title (str): the title of the button dialog
            text (str): the content of the button dialog
            choices (list of choices): each choice can be a string containing the label and the answer a tuple or
                list of two strings where the first entry is the label and the second label is the answer.
            timeout=300 (float): the timeout until the button dialog disappears in seconds (``None`` means no timeout)

        Returns:
            object or None: the answer for the clicked button or ``None`` in case of a timeout
        """
        choices = self._expand_choices(choices)
        ret = self._interaction_server.button_dialog(title=title, text=text, choices=choices, timeout=timeout)
        step = {
            "title": title,
            "text": self._interaction_server.text_to_html(text),
            "type": "button_dialog",
            "choices": choices,
            "reason": None,
            "result": ret,
            "steps": []
        }
        self._result._add_step(step)
        return ret

    def radio_button_dialog(self, title, text, choices, timeout=300):
        """
        Create a dialog with definable radio buttons to select one answer out of many choices.

        Args:
            title (str): the title of the radio button dialog
            text (str): the content of the radio button dialog
            choices (list of choices): each choice can be a string containing the label and the answer a tuple or
                list of two strings where the first entry is the label and the second label is the answer.
                You can also supply a list or tuple with two entries where the first entry is the label and the second
                entry is a boolean that states if the radio button is checked.
                You can also supply a list or tuple with three entries where the first entry is the label,
                the second entry is the answer and the third entry is a boolean that states if the radio button is
                checked.
            timeout=300 (float): the timeout until the radio button dialog disappears in seconds (``None`` means no timeout)

        Returns:
            str or None: the answer for the selected radio button or ``None`` in case of a timeout
        """
        choices = self._expand_choices(choices, enable_checked=True)
        ret = self._interaction_server.radio_button_dialog(title=title, text=text, choices=choices, timeout=timeout)
        step = {
            "title": title,
            "text": self._interaction_server.text_to_html(text),
            "type": "radio_button_dialog",
            "choices": choices,
            "reason": None,
            "result": ret,
            "steps": []
        }
        self._result._add_step(step)
        return ret

    def checkbox_dialog(self, title, text, choices, timeout=300):
        """
        Create a dialog with definable checkboxes for multiple choices.

        Args:
            title (str): the title of the checkbox dialog
            text (str): the content of the checkbox dialog
            choices (list of choices): each choice can be a string containing the label and the answer a tuple or
                list of two strings where the first entry is the label and the second label is the answer.
                You can also supply a list or tuple with two entries where the first entry is the label and the second
                entry is a boolean that states if the radio button is checked.
                You can also supply a list or tuple with three entries where the first entry is the label,
                the second entry is the answer and the third entry is a boolean that states if the radio button is
                checked.
            timeout=300 (float): the timeout until the checkbox dialog disappears in seconds (``None`` means no timeout)

        Returns:
            list of str or None: a list containing the selected answers or ``None`` in case of a timeout
        """
        choices = self._expand_choices(choices, enable_checked=True)
        ret = self._interaction_server.checkbox_dialog(title=title, text=text, choices=choices, timeout=timeout)
        step = {
            "title": title,
            "text": self._interaction_server.text_to_html(text),
            "type": "checkbox_dialog",
            "choices": choices,
            "reason": None,
            "result": ret,
            "steps": []
        }
        self._result._add_step(step)
        return ret

    def file_upload_dialog(self, title, text, file_types=None, timeout=300):
        """
        Create a dialog which allows the user to upload a file.

        Args:
            title (str): the title of the file upload dialog
            text (str): the content of the file upload dialog
            file_types (str): desired file type of files to be uploaded. This should be respected by the file chooser
                in all modern browsers. File types can be specified as file extensions or MIME types, with every
                entry separated by a comma.
            timeout=300 (float): the timeout until the file upload dialog disappears in seconds
                (``None`` means no timeout)

        Returns:
            tuple of str or tuple of None. ``( file name, file type, uploaded file in binary form)``.
        """
        ret = self._interaction_server.file_upload_dialog(title=title, text=text, timeout=timeout,
                                                          file_types=file_types)

        File = namedtuple("File", ("filename", "mime_type", "bytes"))

        if ret is not None:
            uploaded_file = File(filename=ret["name"], mime_type=ret["type"], bytes=base64.b64decode(ret["url"].split(",")[-1]))
        else:
            uploaded_file = [None]

        step = {
            "title": title,
            "text": self._interaction_server.text_to_html(text),
            "type": "file_upload_dialog",
            "reason": None,
            "file_types": file_types,
            "result": uploaded_file[0],
            "steps": []
        }
        self._result._add_step(step)

        return uploaded_file


@fixture('test', auto=True)
class test(_ClassFixture):
    """
    Information about the running test.
    Scope is ``'test'``.
    """
    def __init__(self, result):
        self._result = result

    @property
    def name(self):
        return self._result._test.get_name()

    @property
    def class_name(self):
        return self._result._test.get_class_name()

    @property
    def module(self):
        return self._result._test.get_module_name()

    @property
    def doc(self):
        return self._result._test.get_doc()

    @property
    def tags(self):
        return self._result._test.get_tags()

    @property
    def stdout(self):
        return self._result.get_stdout()

    @property
    def stderr(self):
        return self._result.get_stderr()


@fixture('test', auto=True)
class feature(_ClassFixture):
    """
    Information about the running feature in behavior driven testing.
    Scope is ``'test'``.
    """
    def __init__(self, result):
        self._result = result

    @property
    def name(self):
        if self._result.feature is not None:  # pragma: no branch
            return self._result.feature.name
        else:
            return None  # pragma: no cover

    @property
    def doc(self):
        if self._result.feature is not None:  # pragma: no branch
            return self._result.feature.doc
        else:
            return None  # pragma: no cover

    @property
    def filename(self):
        if self._result.feature is not None:  # pragma: no branch
            return self._result.feature.filename
        else:
            return None  # pragma: no cover


@fixture('test', auto=True)
class scenario(_ClassFixture):
    """
    Information about the running scenario in behavior driven testing.
    Scope is ``'test'``.
    """

    def __init__(self, result):
        self._result = result

    @property
    def name(self):
        if self._result.scenario is not None:  # pragma: no branch
            return self._result.scenario.name
        else:
            return None  # pragma: no cover



@fixture('session', auto=True)
class statistics(_ClassFixture):
    """
    Statistics about the test session.
    Scope is ``'session'``.
    """
    def __init__(self, statistics_provider):
        self._statistics_provider = statistics_provider

    @property
    def number_of_tests(self):
        return self._statistics_provider.number_of_tests

    @property
    def tests_run(self):
        return self._statistics_provider.tests_run

    @property
    def done_percent(self):
        return self._statistics_provider.done_percent

    @property
    def successes(self):
        return self._statistics_provider.successes

    @property
    def failures(self):
        return self._statistics_provider.failures

    @property
    def errors(self):
        return self._statistics_provider.errors

    @property
    def skipped(self):
        return self._statistics_provider.skipped

    @property
    def success_percent(self):
        return self._statistics_provider.success_percent

    @property
    def failure_percent(self):
        return self._statistics_provider.failure_percent

    @property
    def error_percent(self):
        return self._statistics_provider.error_percent

    @property
    def skipped_percent(self):
        return self._statistics_provider.skipped_percent


def _create_flat_query_dict(d, separator="."):
    def _get_flat_dict(data, prefix=""):
        ret = dict()

        for key, value in data.items():
            if not key.startswith("_"):  # pragma: no branch
                if isinstance(value, dict):
                    ret.update(_get_flat_dict(value, prefix=prefix+key+separator))
                else:
                    ret["{}{}".format(prefix, key)] = value

        return ret

    flat = _get_flat_dict(d)
    return flat


def jira_issue_collector(jira, test, result):
    logger = logging.getLogger("htf.jira")
    replacements = {
        "test": test.name,
        "testcase": test.class_name,
        "project": jira._HTF_JIRA_PROJECT
    }

    query = jira._HTF_JIRA_QUERY_TEMPLATE.format(**replacements)

    assert jira._HTF_JIRA_AUTH in [None, "auth", "basic_auth", "oauth", "kerberos"]

    logger.info("Query JIRA issues")

    options = {'server': jira._HTF_JIRA_URL}
    logger.debug("Using JIRA server {}".format(jira._HTF_JIRA_URL))

    if jira._HTF_JIRA_AUTH in ["auth", "basic_auth"]:  # pragma: no cover
        options[jira._HTF_JIRA_AUTH] = (jira._HTF_JIRA_USERNAME, jira._HTF_JIRA_PASSWORD)
        logger.debug("Using authentication scheme {} and username {}".format(jira._HTF_JIRA_AUTH,
                                                                             jira._HTF_JIRA_USERNAME))

    elif jira._HTF_JIRA_AUTH in ["oauth"]:  # pragma: no cover
        options["oauth"] = jira._HTF_JIRA_OAUTH_DICT
        logger.debug("Using authentication scheme {}".format(jira._HTF_JIRA_AUTH))

    elif jira._HTF_JIRA_AUTH in ["kerberos"]:  # pragma: no cover
        options["kerberos"] = True
        options["kerberos_options"] = jira._HTF_JIRA_KERBEROS_OPTIONS
        logger.debug("Using authentication scheme {}".format(jira._HTF_JIRA_AUTH))

    jira_api = jira_module.JIRA(**options)

    # search issues
    logger.debug("query: {}".format(query))
    issues = jira_api.search_issues(query)

    for issue in sorted(issues, key=lambda issue: issue.key):
        flat_issue = _create_flat_query_dict(issue.raw, "/")
        logger.debug("raw issue: {}".format(json.dumps(flat_issue, indent=4, sort_keys=True)))

        result.links.append({'title': jira._HTF_JIRA_TITLE_TEMPLATE.format(**flat_issue),
                             'href': "{}/browse/{}".format(jira._HTF_JIRA_URL, issue.key)})


@fixture('session', auto=True)
class jira(_ClassFixture):
    _HTF_JIRA_URL = None
    _HTF_JIRA_PROJECT = None
    _HTF_JIRA_USERNAME = None
    _HTF_JIRA_PASSWORD = None
    _HTF_JIRA_AUTH = None
    _HTF_JIRA_OAUTH_DICT = None
    _HTF_JIRA_KERBEROS_OPTIONS = None
    _HTF_JIRA_QUERY_TEMPLATE = 'project={project} and (summary~"{testcase}.{test}" or description~"{testcase}.{test}")'
    _HTF_JIRA_TITLE_TEMPLATE = "{fields/issuetype/name} {key} ({fields/status/name}): {fields/summary}"

    def __init__(self, environment):
        self.environment = environment

    def setup_jira_connector(self, url, project,
                             username=None, password=None, auth=None,
                             oauth_dict=None, kerberos_options=None,
                             query_template=None,
                             title_template=None):
        """
        Set up the ``jira`` fixture so that it will search your JIRA installation
        for issues for the current test.

        Args:
            url (str): the url to your JIRA installation
            project (str): the JIRA project
            username=None (str or None): the username to be used for authentication
            password=None (str or None): the password to be used for authentication
            auth=None (str or None): possible values are:

                - ``None``: no authentication is used
                - ``auth``: cookie based authentication
                - ``basic_auth``: HTTP basic authentication
                - ``oauth``: OAuth authentication is used (you have to supply ``oauth_dict``, too)
                - ``kerberos``: Kerberos authentication (you have to supply ``kerberos_options``, too)

            oauth_dict=None (dict): options for ``oauth`` authentication
            kerberos_options=None (dict): options for ``kerberos`` authentication

            query_template="default" (str): the template for the JIRA query string. supported replacements are
                - ``{testcase}``: is replaced with the TestCase's class name
                - ``{test}``: is replaced with the current test name
                - ``{project}``: is replaced with the supplied project

                default: "project={project} and (summary~\"{testcase}.{test}\" or description~\"{testcase}.{test}\")"
            title_template="default" (str): The template for the resulting title of the link in the test report.
                The issue is flattened using ``'/'`` as the separator. So the template may include keys like
                ``{fields/status/name}`` to show the name of the current ticket status, etc.
        """

        self._HTF_JIRA_URL = url
        self._HTF_JIRA_PROJECT = project
        self._HTF_JIRA_USERNAME = username
        self._HTF_JIRA_PASSWORD = password
        self._HTF_JIRA_AUTH = auth
        self._HTF_JIRA_OAUTH_DICT = oauth_dict
        self._HTF_JIRA_KERBEROS_OPTIONS = kerberos_options

        if query_template is not None:  # pragma: no branch
            self._HTF_JIRA_QUERY_TEMPLATE = query_template

        if title_template is not None:  # pragma: no cover
            self._HTF_JIRA_TITLE_TEMPLATE = title_template

        # register jira issue collector
        self.environment.set_fixture("jira_issue_collector", fixture('test', auto=True)(jira_issue_collector))
