# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#
from htf.gherkin.step_parsers import get_parser


def feature(pattern, parser=None, extra_types=None, converters=None):
    """
    Decorates a method to make it an implementation of a given feature in a behavior driven test.

    It can be implemented as a normal method with a return value to implement
    set up code or as a generator to realize set up and tear down code for a feature.

    Args:
        pattern (regex): gherkin feature pattern
        parser=None (condition parser): the parser to parse the parameters from the precondition
        extra_types=None (dict of type converters for parse or parse-type): extra types for parsing
        converters=None (dict with callables or None): optional converters that are applied to
            parameter after parsing
    """
    parser_class = get_parser(parser)

    def wrapper(func):
        func._htf_feature = True
        if not hasattr(func, "_htf_features"):  # pragma: no branch
            func._htf_features = set()

        parser = parser_class(expression=pattern, converters=converters, extra_types=extra_types)
        func._htf_features.add(parser)

        return func

    return wrapper


def scenario(pattern, parser=None, extra_types=None, converters=None):
    """
    Decorates a method to make it an implementation of a given scenario in a behavior driven test.

    It can be implemented as a normal method with a return value to implement
    set up code or as a generator to realize set up and tear down code for a scenario.

    Args:
        pattern (regex): gherkin feature pattern
        parser=None (condition parser): the parser to parse the parameters from the precondition
        extra_types=None (dict of type converters for parse or parse-type): extra types for parsing
        converters=None (dict with callables or None): optional converters that are applied to
            parameters after parsing
    """
    parser_class = get_parser(parser)

    def wrapper(func):
        func._htf_scenario = True
        if not hasattr(func, "_htf_scenarios"):  # pragma: no branch
            func._htf_scenarios = set()

        parser = parser_class(expression=pattern, converters=converters, extra_types=extra_types)
        func._htf_scenarios.add(parser)

        return func

    return wrapper


def given(precondition, parser=None, extra_types=None, converters=None):
    """
    Decorates a method to make it an implementation of a given step in a behavior driven test.
    See also https://cucumber.io/docs/gherkin/reference/#given .

    Args:
        precondition (regex): gherkin feature precondition
        parser=None (condition parser): the parser to parse the parameters from the precondition
        extra_types=None (dict of type converters for parse or parse-type): extra types for parsing
        converters=None (dict with callables or None): optional converters that are applied to
            parameters after parsing
    """
    parser_class = get_parser(parser)

    def wrapper(func):
        func._htf_given = True
        if not hasattr(func, "_htf_given_preconditions"):
            func._htf_given_preconditions = set()

        parser = parser_class(expression=precondition, converters=converters, extra_types=extra_types)
        func._htf_given_preconditions.add(parser)

        return func

    return wrapper


def when(condition, parser=None, extra_types=None, converters=None):
    """
    Decorates a method to make it an implementation of a when step in a behavior driven test.
    See also https://cucumber.io/docs/gherkin/reference/#when .

    Args:
        condition (regex): gherkin feature condition
        parser=None (condition parser): the parser to parse the parameters from the condition
        extra_types=None (dict of type converters for parse or parse-type): extra types for parsing
        converters=None (dict with callables or None): optional converters that are applied to
            parameters after parsing
    """
    parser_class = get_parser(parser)

    def wrapper(func):
        func._htf_when = True
        if not hasattr(func, "_htf_when_conditions"):
            func._htf_when_conditions = set()

        parser = parser_class(expression=condition, converters=converters, extra_types=extra_types)
        func._htf_when_conditions.add(parser)

        return func

    return wrapper


def then(assertion, parser=None, extra_types=None, converters=None):
    """
    Decorates a method to make it an implementation of a then step in a behavior driven test.
    See also https://cucumber.io/docs/gherkin/reference/#then .

    Args:
        assertion (regex): gherkin feature assertion
        parser=None (condition parser): the parser to parse the parameters from the assertion
        extra_types=None (dict of type converters for parse or parse-type): extra types for conversion
        converters=None (dict with callables or None): optional converters that are applied to
            parameters after parsing
    """

    parser_class = get_parser(parser)

    def wrapper(func):
        func._htf_then = True
        if not hasattr(func, "_htf_then_assertions"):
            func._htf_then_assertions = set()

        parser = parser_class(expression=assertion, converters=converters, extra_types=extra_types)
        func._htf_then_assertions.add(parser)

        return func

    return wrapper


def transformation(pattern, parser=None):
    """
    Decorates a method to make it an implementation of a data transformation hook
    used for examples in a behavior driven test.

    Args:
        pattern (regex): gherkin examples pattern
        parser=None (parser): the parser to match the name of the example
    """

    parser_class = get_parser(parser)

    def wrapper(func):
        func._htf_transformation = True
        if not hasattr(func, "_htf_transformations"):  # pragma: no branch
            func._htf_transformations = set()

        parser = parser_class(expression=pattern)
        func._htf_transformations.add(parser)

        return func

    return wrapper


def data_generator(pattern, parser=None):
    """
    Decorates a method to make it an implementation of a data generator
    used for examples in a behavior driven test.

    Args:
        pattern (str): the pattern to match the title
        parser=None (parser): the parser to match the name of the example
    """

    parser_class = get_parser(parser)

    def wrapper(func):
        func._htf_data_generator = True
        if not hasattr(func, "_htf_data_generators"):  # pragma: no branch
            func._htf_data_generators = set()

        parser = parser_class(expression=pattern)
        func._htf_data_generators.add(parser)

        return func

    return wrapper
