# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import os
import re

from htf.util.data_driven_testing import CSVFileIterator, JSONFileIterator, YAMLFileIterator
from htf.gherkin import lang_support
from htf.gherkin.feature import Feature
from htf.gherkin.scenario import Scenario
from htf.gherkin.scenario_outline import ScenarioOutline, DataGenerator, PythonDataGenerator
from htf.gherkin.data_table import DataTable
from htf.gherkin.step import Step
from htf.gherkin.context import Context

indentation_pattern = re.compile(r"^[ \t]*")
language_line_pattern = re.compile(r"^# +language: +([a-z-]+)", re.IGNORECASE)


class GherkinFeatureParser:

    def __init__(self, filename):
        self.filename = filename
        self.lines = None
        self.current_line = 0
        self.dictionary = lang_support.languages["en"]

    def set_language(self, language):
        self.dictionary = lang_support.languages[language]

    def get_indentation(self, line):
        """
        Returns indentation level of the line.
        Only spaces are supported. If a tab is found an exception is raised.

        Args:
            line (str): string of a line

        Returns:
            int: indentation level of line
        """
        match = indentation_pattern.match(line).group(0)
        if "\t" in match:
            self.raise_syntax_error("A tab (\\t) was found as an indentation character. Only spaces are supported!",
                                    self.current_line)
        return len(match)

    def is_ignored(self, line):
        """
        Check if a line is ignored.

        Args:
            line (str): the line to be checked

        Returns:
            bool: ``True`` is the line is ignored
        """
        if line.strip() == "" or line.strip().startswith("#"):
            return True
        return False

    def match_keyword_return_rest(self, keyword, line):
        """
        Match a given keyword taken from the current language dictionary and return the rest.

        Args:
            keyword (str): the keyword to be looked up
            line (str): the current line

        Returns:
            str: the part of the line behind the keyword
        """
        for alias in self.dictionary[keyword]:
            if keyword in ["given", "when", "then", "but", "and"]:
                found = line.find(alias)
                if found != - 1:
                    return line[found + len(alias):].strip()
            else:
                found = line.find(alias + ":")
                if found != -1:
                    padding_count = 2
                    start_index = found + len(alias) + padding_count
                    return line[start_index:].strip()
        return None

    def generate_lines(self):
        """
        Generates tuples of not ignored lines and line numbers and the level of indentation.
        """
        with open(self.filename, encoding="utf-8") as reader:
            lines = reader.readlines()

            continued_line = None  # for escaped line
            continued_indentation = None

            for line_number, line in enumerate(lines, start=1):
                # check for language
                if line_number == 1:
                    match = language_line_pattern.match(line)
                    if match:
                        language = match.group(1)
                        if language not in lang_support.languages:
                            raise Exception("language '{}' is not supported in file: {}:{}".format(
                                language, self.filename, line_number))
                        self.set_language(language)
                        continue

                # is line ignored?
                if continued_line is None and self.is_ignored(line):
                    continue
                
                # get indentation
                indentation = self.get_indentation(line)

                # strip line
                line = line.strip().replace("\r", "").replace("\n", "")

                # is the line continued?
                if line.endswith("\\"):
                    line = line[:-1].strip()  # remove \ at the end
                    # construct continued line
                    if continued_line is None:
                        continued_line = line
                    else:
                        continued_line += " " + line
                    # store continued indentation for the first continued line
                    if continued_indentation is None:
                        continued_indentation = indentation
                else:
                    if continued_line is not None:
                        # a continued line ends here
                        continued_line += " " + line
                        yield continued_line, line_number, continued_indentation
                        # restore
                        continued_line = None
                        continued_indentation = None
                    else:
                        # yield
                        yield line, line_number, indentation

    def get_next_line(self):
        try:
            return self.lines[self.current_line]
        except IndexError:
            # last line was parsed
            return None, None, None

    def consume_line(self):
        self.current_line += 1

    def parse_tags_and_requirements_and_skip(self, container, indentation):
        """
        Try to parse ``Tags:``, ``Requirements:`` and ``Skip:``and append them to ``container``.

        Args:
            container: a Feature, Scenario or ScenarioOutline
            indentation: the line indentation of the enclosing block
        """
        for i in range(3):
            line, line_number, line_indentation = self.get_next_line()
            if line is None:
                return None

            if line.startswith("Requirements:") or line.startswith("Tags:") or line.startswith("Skip:"):
                if line_indentation <= indentation:
                    self.raise_syntax_error("Requirements and Tags indentation heading has to be indented!", line_number)

                if line.startswith("Requirements:"):
                    # parse requirements
                    requirements = self._get_keyword_value("requirements", line)
                    requirements = list(sorted([r.strip() for r in requirements.split(",")]))
                    container.set_requirements(requirements)

                elif line.startswith("Tags:"):  # pragma: no branch
                    tags = self._get_keyword_value("tags", line)
                    tags = list(sorted([r.strip() for r in tags.split(",")]))
                    container.set_tags(tags)

                elif line.startswith("Skip:"):  # pragma: no branch
                    skipped_reason = self._get_keyword_value("skip", line)
                    container.set_skipped(skipped_reason)

                self.consume_line()

    def parse_feature(self):
        """
        Parse a feature starting from the current line.
        """
        line, line_number, indentation = self.get_next_line()
        if line is None:
            return None

        # line has to start with "Feature: (translated)"
        feature_name = self.match_keyword_return_rest("feature", line)
        if feature_name is None:
            self.raise_syntax_error( "Cannot read .feature file, invalid format", line_number)
        self.consume_line()

        info = self.parse_info(indentation=indentation)
        feature = Feature(name=feature_name, doc=info, filename=self.filename, line_number=line_number)

        # Try to parse tags, requirements and skip
        self.parse_tags_and_requirements_and_skip(feature, indentation=indentation)

        feature.background = self.parse_background(indentation=indentation)

        change = False
        while True:
            # parse scenarios
            scenarios = self.parse_scenarios(indentation=indentation)
            if len(scenarios) > 0:
                change = True
            for scenario in scenarios:
                feature.add_scenario(scenario)

            # parse scenario outlines
            scenario_outlines = self.parse_scenario_outlines(indentation=indentation)
            if len(scenario_outlines) > 0:
                change = True
            for scenario_outline in scenario_outlines:
                feature.add_scenario_outline(scenario_outline)

            # parse rule
            rule_scenarios = self.parse_rule(indentation=indentation)
            if len(rule_scenarios) > 0:
                change = True
            for rule_scenario in rule_scenarios:
                feature.add_scenario(rule_scenario)

            if len(feature.scenarios) + len(feature.scenario_outlines) < 1:
                self.raise_syntax_error("A feature must contain at least one scenario, scenario outline or rule!",
                                        line_number)
            if not change:
                break
            change = False

        return feature

    def parse_info(self, indentation):
        """
        Read info by tracking indentation and line number changes.
        """
        info = ""
        line, line_number, line_indentation = self.get_next_line()
        if line is None:
            return info

        def contains_any_keyword(self, line):
            keywords = ["background", "examples", "feature", "rule", "scenario", "scenario_outline", "given", "when",
                        "then", "and", "but"]
            for keyword in keywords:
                if self.match_keyword_return_rest(keyword, line) is not None:
                    return True
            if line.startswith("Requirements:") or line.startswith("Tags:") or line.startswith("Skip:"):
                return True

            return False

        if contains_any_keyword(self, line):
            return info

        last_line_number = line_number
        last_line_indentation = line_indentation
        info += line

        self.consume_line()

        if last_line_indentation <= indentation:
            raise self.raise_syntax_error("Info string has to be indented!", line_number)

        while True:
            line, line_number, line_indentation = self.get_next_line()
            if line is None:
                return info
            if contains_any_keyword(self, line):
                return info
            diff = line_number - last_line_number
            if last_line_indentation == line_indentation and diff == 1:
                info += "\n" + line
                last_line_number = line_number
                last_line_indentation = line_indentation
                self.consume_line()
            else:
                return info

    def parse_background(self, indentation):
        """
        Read background by tracking indentation and line number changes.
        """
        background = []
        line, line_number, line_indentation = self.get_next_line()
        if line is None:
            return background

        # first line should start with "Background: "
        background_name = self.match_keyword_return_rest("background", line)
        if background_name is None:
            return background

        if line_indentation <= indentation:
            self.raise_syntax_error("Background heading has to be indented!", line_number)
        self.consume_line()

        # Parse info
        info = self.parse_info(indentation=line_indentation)  # currently has no effect

        # parse Given
        givens = self.parse_steps(keyword="given", indentation=line_indentation)
        for given in givens:
            background.append(given)

        return background

    def parse_rule(self, indentation):
        """
        Read rule , all contained scenarios will be part of this rule
        Returns: List of scenarios
        """
        line, line_number, line_indentation = self.get_next_line()
        if line is None:
            return []

        # first line should start with "Rule: "
        rule_name = self.match_keyword_return_rest("rule", line)
        if rule_name is None:
            return []

        if line_indentation <= indentation:
            self.raise_syntax_error("Rule heading has to be indented!", line_number)
        self.consume_line()

        # Parse info
        info = self.parse_info(indentation=line_indentation)  # currently has no effect

        rule_scenarios = self.parse_scenarios(line_indentation)
        for rule_scenario in rule_scenarios:
            rule_scenario.rule = rule_name
        return rule_scenarios

    def parse_scenarios(self, indentation):
        """
        Read scenarios by tracking indentation and line number changes.
        """
        scenarios = []

        while True:
            scenario = self.parse_scenario(indentation=indentation)
            if not scenario:
                break
            scenarios.append(scenario)

        return scenarios

    def parse_scenario(self, indentation):
        """
        Read next scenario by tracking indentation and line number changes.
        """
        line, line_number, line_indentation = self.get_next_line()
        if line is None:
            return None

        # first line should start with "Scenario: "
        scenario_name = self.match_keyword_return_rest("scenario", line)
        if scenario_name is None:
            return None

        if line_indentation <= indentation:
            return None

        self.consume_line()

        info = self.parse_info(indentation=line_indentation)
        scenario = Scenario(name=scenario_name, doc=info, filename=self.filename, line_number=line_number)
        # Try to parse tags and requirements
        self.parse_tags_and_requirements_and_skip(scenario, indentation=indentation)

        # parse Given
        givens = self.parse_steps(keyword="given", indentation=line_indentation)
        for given in givens:
            scenario.add_given(given)

        # parse Whens
        whens = self.parse_steps(keyword="when", indentation=line_indentation)
        for when in whens:
            scenario.add_when(when)

        # parser Thens
        thens = self.parse_steps(keyword="then", indentation=line_indentation)
        for then in thens:
            scenario.add_then(then)

        return scenario

    def parse_steps(self, keyword, indentation):
        """
        Parse a block of steps starting with ``keyword`` and include ``Ands`` and ``Buts``.
        """
        steps = []

        while True:
            line, line_number, line_indentation = self.get_next_line()
            if line is None:
                return steps

            if line_indentation <= indentation:
                # block is over
                break

            # check for keyword
            keyword_step = self.match_keyword_return_rest(keyword, line)
            if keyword_step is not None:  # pragma: no branch
                step = Step(expression=keyword_step, line=line, filename=self.filename, line_number=line_number)
                self.consume_line()
                step.set_context(self.parse_context(line_indentation))
                steps.append(
                    step
                )

                # check for "And"
                while True:
                    # check for and and but steps following the keyword
                    line, line_number, line_indentation = self.get_next_line()

                    if line is None:
                        return steps

                    if line_indentation <= indentation:
                        # block is over
                        break

                    and_step = self.match_keyword_return_rest("and", line)
                    if and_step is not None:
                        step = Step(expression=and_step, line=line, filename=self.filename, line_number=line_number)
                        self.consume_line()
                        step.set_context(self.parse_context(line_indentation))
                        steps.append(
                            step
                        )
                    else:
                        # check for "But"
                        but_step = self.match_keyword_return_rest("but", line)
                        if but_step is not None:  # pragma: no branch
                            step = Step(expression=but_step, line=line, filename=self.filename,
                                        line_number=line_number)
                            self.consume_line()
                            step.set_context(self.parse_context(line_indentation))
                            steps.append(
                                step
                            )
                        else:  # pragma: no cover
                            # no step was consumed
                            break
            else:  # pragma: no cover
                break

        return steps

    def parse_examples(self, scenario_outline, indentation):
        """
        Parse one "Examples:" section and add the data generator to ``scenario_outline``.
        """
        line, line_number, examples_line_indentation = self.get_next_line()
        if line is None:
            return None

        # next line should start with "Examples: "
        examples_name = self.match_keyword_return_rest("examples", line)

        # set examples name
        scenario_outline.set_examples_name(examples_name)

        if examples_line_indentation <= indentation:
            self.raise_syntax_error("Examples heading has to be indented!", line_number)

        self.consume_line()

        # parse example data lines
        while True:
            data_generator = self.parse_data_generator(examples_line_indentation, expect_data_source=True)

            if data_generator is not None:
                scenario_outline.add_data_generator(data_generator)
            else:  # pragma: no cover
                break

        return True

    def parse_scenario_outlines(self, indentation):
        """
        Read scenario outlines by tracking indentation and line number changes.
        
        Args:
            indentation (int): the outer indentation
        """
        scenario_outlines = []

        while True:
            scenario_outline = self.parse_scenario_outline(indentation=indentation)
            if not scenario_outline:
                break
            scenario_outlines.append(scenario_outline)

        return scenario_outlines

    def parse_scenario_outline(self, indentation):
        """
        Read next scenario outline by tracking indentation and line number changes.
        """
        line, line_number, line_indentation = self.get_next_line()
        if line is None:
            return None

        # first line should start with "Scenario Outline: "
        scenario_outline_name = self.match_keyword_return_rest("scenario_outline", line)
        if scenario_outline_name is None:
            return None

        if line_indentation <= indentation:
            self.raise_syntax_error("Scenario Outline heading has to be indented!", line_number)

        self.consume_line()
        # Parse info
        info = self.parse_info(indentation=line_indentation)
        scenario_outline = ScenarioOutline(name=scenario_outline_name, doc=info, filename=self.filename,
                                           line_number=line_number)
        # Try to parse tags and requirements
        self.parse_tags_and_requirements_and_skip(scenario_outline, indentation=line_indentation)

        # parse Given
        givens = self.parse_steps(keyword="given", indentation=line_indentation)
        for given in givens:
            scenario_outline.add_given(given)

        # parse Whens
        whens = self.parse_steps(keyword="when", indentation=line_indentation)
        for when in whens:
            scenario_outline.add_when(when)

        # parse Thens
        thens = self.parse_steps(keyword="then", indentation=line_indentation)
        for then in thens:
            scenario_outline.add_then(then)

        # parse Examples
        self.parse_examples(scenario_outline, indentation=line_indentation)

        return scenario_outline

    def parse_context(self, indentation):
        context = Context()
        table = self.parse_data_generator(indentation)
        if table:
            context.set_table(table)
            text = self.parse_text(indentation)
            if text:
                context.set_text(text)
        else:
            text = self.parse_text(indentation)
            if text:
                context.set_text(text)
                table = self.parse_data_generator(indentation)
                if table:
                    context.set_table(table)
        return context

    def parse_text(self, indentation):
        """
        Parse a block of text, which will be passed into context of a step
        """
        line, line_number, line_indentation = self.get_next_line()
        if line is None or not line.startswith('"""'):
            return None

        if line_indentation <= indentation:
            self.raise_syntax_error("step context text should be indented!", line_number)

        line = line[3:]  # cut """ from beginning
        text = line  # add first text line
        self.consume_line()

        while not text.endswith('"""'):
            # Read value rows until text ends
            line, line_number, line_indentation = self.get_next_line()
            if line is None or line_indentation <= indentation:
                self.raise_syntax_error('Text block is not closed properly, use """', line_number)

            text = text + line
            self.consume_line()

        return text[:len(text)-3]  # remove """ from end

    def _get_filename_from_data_generator(self, keyword, line, directory):
        """
        Return the filename for a data generator line
        """
        filename = self._get_keyword_value(keyword, line)
        path = os.path.normpath(os.path.join(directory, filename))

        return path

    def _get_title_from_data_generator(self, keyword, line):
        """
        Return the title for a data generator
        """
        return self._get_keyword_value(keyword, line)

    def _get_keyword_value(self, keyword, line):
        """
        Return the keyword for a data generator line
        """
        assert line.lower().startswith(keyword.lower() + ":")
        value = line[len(keyword)+1:].strip()

        return value

    def parse_data_generator(self, indentation, expect_data_source=False):
        """
         Parse a block of data table or external data sources which will be cross-tested in the scenario outline,
         or used as a parameter in a step.

         Returns:
             iterable: containing data sources dictionaries
         """

        # get first line containing the keys
        line, line_number, line_indentation = self.get_next_line()
        if line is None:
            return None

        # get scenario outline directory
        directory = os.path.dirname(self.filename)

        if line.startswith("CSV:"):
            if line_indentation <= indentation:
                self.raise_syntax_error("CSV data source should be indented!", line_number)
            self.consume_line()
            filename = self._get_filename_from_data_generator("csv", line, directory)
            csv_file_iterator = DataGenerator(CSVFileIterator(filename), self.filename, line_number=line_number)
            return csv_file_iterator
        elif line.startswith("JSON:"):
            if line_indentation <= indentation:
                self.raise_syntax_error("JSON data source should be indented!", line_number)
            self.consume_line()
            filename = self._get_filename_from_data_generator("json", line, directory)
            json_file_iterator = DataGenerator(JSONFileIterator(filename), self.filename, line_number=line_number)
            return json_file_iterator
        elif line.startswith("YAML:"):
            if line_indentation <= indentation:
                self.raise_syntax_error("YAML data source should be indented!", line_number)
            self.consume_line()
            filename = self._get_filename_from_data_generator("yaml", line, directory)
            yaml_file_iterator = DataGenerator(YAMLFileIterator(filename), self.filename, line_number=line_number)
            return yaml_file_iterator
        elif line.startswith("Generator:"):
            if line_indentation <= indentation:
                self.raise_syntax_error("Generator data source should be indented!", line_number)
            self.consume_line()
            title = self._get_title_from_data_generator("generator", line)
            data_generator = PythonDataGenerator(title, self.filename, line_number=line_number)
            return data_generator
        elif line.startswith("|"):
            # read a data table
            if line_indentation <= indentation:
                self.raise_syntax_error("data table should be indented!", line_number)

            def data_line_splitter(line):
                """
                Takes the data line and splits it into meaningful data list
                Args:
                    line: data line to interpret
                Returns:
                    A tuple of : (A list containing values of the line , A list containing positions of splitters)
                """
                line_splitted = line.split("|")
                line_splitted = line_splitted[1:len(line_splitted) - 1]
                headers = [data.strip() for data in line_splitted]

                def find_offsets(haystack, needle):
                    """
                    Find the start of all (possibly-overlapping) instances of needle in haystack
                    """
                    res = []
                    offs = -1
                    while True:
                        offs = haystack.find(needle, offs + 1)
                        if offs == -1:
                            break
                        else:
                            res.append(offs)
                    return res

                return headers, find_offsets(line, "|")

            data_table = DataTable()

            # Read first data line, which will contain variable names
            headers, separator_positions = data_line_splitter(line)
            data_table.set_headers(headers)
            self.consume_line()
            while True:
                # Read value rows until indentation breaks
                line, line_number, line_indentation = self.get_next_line()
                if line is None:
                    break

                if not line.startswith("|"):
                    # table is over
                    break

                # check line indentation
                if line_indentation <= indentation:
                    self.raise_syntax_error('Table block is not indented properly', line_number)

                data_line, line_separator_positions = data_line_splitter(line)
                if separator_positions != line_separator_positions:
                    self.raise_syntax_error('Table block is not separated properly, make sure to align | symbols',
                                            line_number)
                data_table.add_line(data_line)
                self.consume_line()

            data_table_iterator = DataGenerator(data_table, self.filename, line_number=line_number)

            return data_table_iterator
        else:
            # if the line does not start with a keyword a syntax error is raised
            if expect_data_source is True:
                self.raise_syntax_error("Unknown data source in '{}'".format(line), line_number)

    def raise_syntax_error(self, message, line_number):
        raise SyntaxError("{} ({}:{})".format(message, self.filename, line_number))

    def get_feature_from_file(self):
        """
        Parse the current file and return a Feature object

        Returns:
            Feature object containing all information in its attributes
        """

        # read lines
        self.lines = list(self.generate_lines())

        # parse and return one feature
        feature = self.parse_feature()

        # there shouldn't be any other lines left
        if self.current_line < len(self.lines) - 1:

            def build_original_line(line):
                line, line_number, indentation = line
                original_line = " " * indentation
                original_line += line + " (line: {})".format(line_number)
                return original_line

            raise SyntaxError("There are unparsed lines in the feature file: {}\n".format(self.filename) +
                              "\n".join([build_original_line(line) for line in self.lines[self.current_line:]]))

        return feature
