# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import inspect
import re

indentation_pattern = re.compile(r"^[ \t]*")


class GherkinSteps:
    """
    This class holds gherkin steps.
    """

    def __init__(self):
        self._features = set()
        self._scenarios = set()
        self._given_preconditions = set()
        self._when_conditions = set()
        self._then_assertions = set()
        self._transformations = set()
        self._data_generators = set()

    @staticmethod
    def get_indentation(line):
        match = indentation_pattern.match(line).group(0)
        if match:  # pragma: no cover
            return len(match)
        else:  # pragma: no cover
            return 0

    def get_printable_method(self, method):
        filename = inspect.getsourcefile(method)
        source, line_number = inspect.getsourcelines(method)

        __str__ = "'{}' defined in File \"{}\", line {}:\n".format(method.__name__, filename, line_number)
        if len(source) > 5:
            last_line = source[5]
            source = source[:5] + [" " * self.get_indentation(last_line) + "# %< %< %< (snipped)"]
        __str__ += "".join(source) + "\n"

        return __str__

    def get_matches_printable(self, matches):
        """
        Return all matches in a printable way.
        """
        return "\n".join(self.get_printable_method(match) for match, _ in matches)

    def register_feature(self, implementation):
        """
        Register a @feature.

        Args:
            implementation (callable): implementation of a @feature
        """
        self._features.add(implementation)

    def get_feature(self, pattern, filename, line_number):
        """
        Get the implementation for a @feature.

        Args:
            pattern (str): the current pattern of a feature name
            filename (str): the filename of the current feature
            line_number (int): the line number of the current feature

        Returns:
            None or tuple of (function, parameters)
        """
        matches = []

        for func in self._features:
            # try to match and convert the parameters
            for p in func._htf_features:
                match = p.match(pattern)
                if match is not None:
                    matches.append((func, match))

        if len(matches) == 0:
            return None

        if len(matches) >= 2:
            raise Exception("Multiple @feature implementations found for pattern '{}' in '{}:{}'".format(
                pattern, filename, line_number) + "\n\n" + self.get_matches_printable(matches))

        # we found one match
        return matches[0]

    def register_scenario(self, implementation):
        """
        Register a @scenario.

        Args:
            implementation (callable): implementation of a @scenario
        """
        self._scenarios.add(implementation)

    def get_scenario(self, pattern, filename, line_number):
        """
        Get the implementation for a @scenario.

        Args:
            pattern (str): the current pattern of a scenario name
            filename (str): the filename of the current feature
            line_number (int): the line number of the current scenario

        Returns:
            None or tuple of (function, parameters)
        """
        matches = []

        for func in self._scenarios:
            # try to match and convert the parameters
            for p in func._htf_scenarios:
                match = p.match(pattern)
                if match is not None:
                    matches.append((func, match))

        if len(matches) == 0:
            return None

        if len(matches) >= 2:
            raise Exception("Multiple @scenario implementations found for pattern '{}' in '{}:{}'".format(
                pattern, filename, line_number))

        # we found one match
        return matches[0]

    def register_given_precondition(self, precondition):
        """
        Register a @given precondition.

        Args:
            precondition (callable): implementation of a @given step
        """
        self._given_preconditions.add(precondition)

    def get_given_precondition(self, precondition, filename, line_number):
        """
        Get the implementation for a @given precondition.

        Args:
            precondition (str): the current precondition from a feature file.
            filename (str): the filename of the current feature
            line_number (int): the line number of the current precondition

        Returns:
            tuple of (function, parameters)
        """
        matches = []

        for func in self._given_preconditions:
            # try to match and convert the parameters
            for p in func._htf_given_preconditions:
                match = p.match(precondition)
                if match is not None:
                    matches.append((func, match))

        # assert that exactly one match was found
        if len(matches) == 0:
            raise Exception("No @given implementation found for precondition '{}' in '{}:{}'".format(
                precondition, filename, line_number))

        if len(matches) >= 2:
            raise Exception("Multiple @given implementations found for precondition '{}' in '{}:{}'".format(
                precondition, filename, line_number) + "\n\n" + self.get_matches_printable(matches))

        # we found one match
        return matches[0]

    def register_when_condition(self, condition):
        """
        Register a @when condition.

        Args:
            condition (callable): implementation of a @when step
        """
        self._when_conditions.add(condition)

    def get_when_condition(self, condition, filename, line_number):
        """
        Get the implementation for a @when condition.

        Args:
            condition (str): the current condition from a feature file.
            filename (str): the filename of the current feature
            line_number (int): the line number of the current condition

        Returns:
            tuple of (function, parameters)
        """
        matches = []

        for func in self._when_conditions:
            # try to match and convert the parameters
            for p in func._htf_when_conditions:
                match = p.match(condition)
                if match is not None:
                    matches.append((func, match))

        # assert that exactly one match was found
        if len(matches) == 0:
            raise Exception("No @when implementation found for condition '{}' in '{}:{}'".format(
                condition, filename, line_number))

        if len(matches) >= 2:
            raise Exception("Multiple @when implementations found for condition '{}' in '{}:{}'".format(
                condition, filename, line_number) + "\n\n" + self.get_matches_printable(matches))

        # we found one match
        return matches[0]

    def register_then_assertion(self, assertion):
        """
        Register a @then condition.

        Args:
            assertion (callable): implementation of a @given step
        """
        self._then_assertions.add(assertion)

    def get_then_condition(self, assertion, filename, line_number):
        """
        Get the implementation for a @then assertion.

        Args:
            assertion (str): the current assertion from a feature file.
            filename (str): the filename of the current feature
            line_number (int): the line number of the current assertion
        """
        matches = []

        for func in self._then_assertions:
            # try to match and convert the parameters
            for p in func._htf_then_assertions:
                match = p.match(assertion)
                if match is not None:
                    matches.append((func, match))

        # assert that exactly one match was found
        if len(matches) == 0:
            raise Exception("No @then implementation found for assertion '{}' in '{}:{}'".format(
                assertion, filename, line_number))

        if len(matches) >= 2:
            raise Exception("Multiple @then implementations found for assertion '{}' in '{}:{}'".format(
                assertion, filename, line_number) + "\n\n" + self.get_matches_printable(matches))

        # we found one match
        return matches[0]

    def register_transformation(self, transformer):
        """
        Register a @transformation data transformation handler.

        Args:
            transformer (callable): implementation of a @transformations data transformation handler
        """
        self._transformations.add(transformer)

    def get_transformation(self, name, filename, line_number):
        """
        Get the implementation for a @transformation data transformation handler.

        Args:
            name (str): the current name of the example from a feature file.
            filename (str): the filename of the current feature
            line_number (int): the line number of the current assertion
        """
        matches = []

        for func in self._transformations:
            # try to match and convert the parameters
            for p in func._htf_transformations:
                match = p.match(name)
                if match is not None:
                    matches.append((func, match))

        if len(matches) == 0:
            return None

        if len(matches) >= 2:
            raise Exception("Multiple @transformation implementations found for pattern '{}' in '{}:{}'".format(
                name, filename, line_number) + "\n\n" + self.get_matches_printable(matches))

        # we found one match
        return matches[0][0]  # no parameters needed

    def register_data_generator(self, data_generator):
        """
        Register a @data_generator data generator.

        Args:
            data_generator (callable): implementation of a @data_generator data generator
        """
        self._data_generators.add(data_generator)

    def get_data_generator(self, title, filename, line_number):
        """
        Get the implementation for a @data_generator data generator.

        Args:
            title (str): the current title of the generator from a feature file.
            filename (str): the filename of the current feature
            line_number (int): the line number of the current assertion
        """
        matches = []

        for func in self._data_generators:
            # try to match and convert the parameters
            for p in func._htf_data_generators:
                match = p.match(title)
                if match is not None:  # pragma: no cover
                    matches.append((func, match))

        # assert that exactly one match was found
        if len(matches) == 0:
            raise Exception("No @data_generator implementation found for title '{}' in '{}:{}'".format(
                title, filename, line_number))

        if len(matches) >= 2:
            raise Exception("Multiple @data_generator implementations found for pattern '{}' in '{}:{}'".format(
                title, filename, line_number) + "\n\n" + self.get_matches_printable(matches))

        # we found one match
        return matches[0][0]  # no parameters needed

    def combine(self, other):
        self._features |= other._features
        self._scenarios |= other._scenarios
        self._given_preconditions |= other._given_preconditions
        self._when_conditions |= other._when_conditions
        self._then_assertions |= other._then_assertions
        self._transformations |= other._transformations
        self._data_generators |= other._data_generators
