# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#
from htf.gherkin.scenario import Scenario


class DataGenerator:
    def __init__(self, generator, filename, line_number=None):
        self.generator = generator
        self.filename = filename
        self.line_number = line_number


class PythonDataGenerator:
    def __init__(self, title, filename, line_number=None):
        self.title = title
        self.generator = None
        self.filename = filename
        self.line_number = line_number

    def set_generator(self, generator):
        """
        Search for the data generator and setup generator.
        """
        self.generator = generator()


class ScenarioOutline:
    def __init__(self, name, doc, filename, line_number):
        self.name = name
        self.doc = doc
        self.examples_name = None
        self.transformation = None
        self.filename = filename
        self.line_number = line_number

        self.given_preconditions = list()
        self.when_conditions = list()
        self.then_assertions = list()

        self.data_generators = []

        # Tags, requirements and skip reasons
        self.tags = None
        self.requirements = None
        self.skip_reason = None

    def set_examples_name(self, name):
        self.examples_name = name

    def add_given(self, step):
        self.given_preconditions.append(step)

    def add_when(self, step):
        self.when_conditions.append(step)

    def add_then(self, step):
        self.then_assertions.append(step)

    def add_data_generator(self, data_generator):
        assert data_generator is not None
        self.data_generators.append(data_generator)

    def setup_transformation(self, gherkin_steps):
        self.transformation = gherkin_steps.get_transformation(
            self.examples_name, self.filename, self.line_number)  # no parameters needed here

    def setup_data_generators(self, gherkin_steps):
        for generator in self.data_generators:
            if isinstance(generator, PythonDataGenerator):
                # the generator needs setup
                generator.set_generator(
                    gherkin_steps.get_data_generator(
                        generator.title, generator.filename, generator.line_number)
                )

    def __iter__(self):
        """
        Generate Scenarios with replacements.
        """

        for data_generator in self.data_generators:
            for replacements in data_generator.generator:
                # transform data
                if self.transformation:
                    replacements = self.transformation(replacements)

                scenario = Scenario(name=self.name, doc="", filename=self.filename, line_number=self.line_number,
                                    kwargs=replacements)

                # set tags and requirements
                scenario.set_tags(self.tags)
                scenario.set_requirements(self.requirements)
                scenario.set_skipped(self.skip_reason)

                # deep copy data and replace everything
                for given_precondition in self.given_preconditions:
                    scenario.add_given(given_precondition.replace(replacements, data_generator.filename))
                for when_condition in self.when_conditions:
                    scenario.add_when(when_condition.replace(replacements, data_generator.filename))
                for then_assertion in self.then_assertions:
                    scenario.add_then(then_assertion.replace(replacements, data_generator.filename))

                yield scenario

    def set_tags(self, tags):
        self.tags = tags

    def set_requirements(self, requirements):
        self.requirements = requirements

    def set_skipped(self, reason):
        self.skip_reason = reason
