# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import re
import parse
from parse_type.cfparse import Parser as CFParser


class _BaseGherkinStepParser:
    def __init__(self, expression, converters=None, extra_types=None):
        """
        Base parser for @feature, @scenario, @given, @when and @then implementations.

        It matches a precondition, a condition or an assertion, extracts the parameters
        and optionally applies conversion to them.

        Args:
            expression (regex): an expression that can be a precondition, a condition or an assertion
            converters=None (iterable of callables): converters for the matching parameters
            extra_types=None (dict of type converters for parse or parse-type): extra types for conversion
        """
        self.expression = expression
        self.converters = converters
        self.extra_types = extra_types
        if self.converters is not None:  # pragma: no cover
            assert isinstance(self.converters, dict), "converters must be a dictionary"

    def apply_converters(self, parameters):
        """
        Apply converters to parameters and return the result.
        """
        for key in parameters:
            if self.converters:
                if key in self.converters: # pragma: no branch
                    parameters[key] = self.converters[key](parameters[key])

        return parameters

    def match(self, step):
        """
        Match a step from a feature file.

        Returns:
            None or dictionary of parameters with optional converters applied
        """
        raise NotImplementedError # pragma: no cover

    def __str__(self):  # pragma: no cover
        return "<{} expression='{}' converters={}>".format(self.__class__.__name__, self.expression, self.converters)

    __repr__ = __str__


class GherkinDefaultParser(_BaseGherkinStepParser):
    def __init__(self, expression, converters=None, extra_types=None):
        """
        Simple parser for @feature, @scenario, @given, @when and @then implementations.

        It matches a precondition, a condition or an assertion, extracts the parameters
        and optionally applies converters to them.

        Parameters have to be supplied with names to return a group-dict.

        Most of the Python format specification is supported:

        https://docs.python.org/3/library/string.html#format-specification-mini-language

        Args:
            expression (str): an expression that can be a precondition, a condition or an assertion
                in the format for the ``parse`` packet
            converters=None (iterable of callables): converters for the matching parameters
            extra_types=None (dict of type converters for parse or parse-type): extra types for conversion
        """
        super(GherkinDefaultParser, self).__init__(expression, converters, extra_types)

        self.pattern = self.create_parser(expression)

    def create_parser(self, expression):
        return parse.compile(expression, extra_types=self.extra_types)

    def match(self, step):
        """
        Match a step from a feature file.

        Returns:
            None or dictionary of parameters with optional converters applied
        """
        if self.pattern is None or step is None:  # pragma: no cover
            return None

        matches = self.pattern.parse(step)

        if matches:
            # apply converters
            fixed = matches.fixed
            if len(fixed):
                raise ValueError(r"Use named parameters only: '{name}'")

            parameters = self.apply_converters(matches.named)
            return parameters
        else:
            return None


class GherkinCardinalityFieldParser(GherkinDefaultParser):
    def create_parser(self, expression):
        return CFParser(expression, extra_types=self.extra_types)


class GherkinRegExStepParser(_BaseGherkinStepParser):
    def __init__(self, expression, converters=None, extra_types=None):
        """
        Regular expression parser for @feature, @scenario, @given, @when and @then implementations.

        It matches a precondition, a condition or an assertion, extracts the parameters
        and optionally applies converters to them.

        Parameters have to be supplied with names to return a group-dict.

        Args:
            expression (regex): an expression that can be a precondition, a condition or an assertion
            converters=None (iterable of callables): converters for the matching parameters
        """
        if extra_types:  # pragma: no cover
            raise TypeError("Do not use extra types for regular expression step parser!")

        super(GherkinRegExStepParser, self).__init__(expression, converters)
        # add chars to always match the full line
        self.pattern = re.compile(expression)

    def match(self, step):
        """
        Match a step from a feature file.

        Returns:
            None or dictionary of parameters with optional converters applied
        """
        if self.pattern is None or step is None:  # pragma: no cover
            return None

        matches = self.pattern.fullmatch(step)

        if matches:
            # apply converters
            groups = matches.groups()
            groupdict = matches.groupdict()
            if len(groups) != len(groupdict):
                raise ValueError(r"Use group dict reg-ex expressions: r'(?P<name>\d+)'")
            parameters = self.apply_converters(groupdict)
            return parameters
        else:
            return None


PARSER_MAPPING = {}
DEFAULT_PARSER = "default"


def register_parser(names, parser):
    """
    Register a behavior driven testing step parser.

    Args:
        names (str or list of str): then name of the parser to be used
        parser (class): the parser class
    """
    if not isinstance(names, list):  # pragma: no branch
        names = [names] # pragma: no cover

    for name in names:
        PARSER_MAPPING[name] = parser


def get_parser(parser):
    """
    Return the class of a behavior driven testing parser.
    The default parser is the simple parser based on the ``parse`` packet.

    Args:
        parser (class or str): the name or class of a parser.

    Returns:
        a class of a parser to parse steps.
    """
    if parser is None:
        parser = DEFAULT_PARSER

    if parser in PARSER_MAPPING:
        return PARSER_MAPPING[parser]
    else:
        return parser



# register default available parsers
register_parser(["default", "parse"], GherkinDefaultParser)
register_parser(["cfparse", "cf"], GherkinCardinalityFieldParser)
register_parser(["regex", "re"], GherkinRegExStepParser)
