# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import random


class TestSuite:
    """
    A generator for a collection of tests
    """
    def __init__(self, discovery, tests_filter_func=None, fixtures_filter_func=None, shuffle=False):
        """
        Args:
            discovery (Discovery): a discovery instance
            tests_filter_func (callable): optional filter function for tests
            fixtures_filter_func (callable): optional filter function for fixtures
            shuffle=False (bool): set to ``True`` to shuffle tests
        """
        self._tests = discovery.discover_tests()
        self._fixtures = discovery.discover_fixtures()
        self._tests_filter_func = tests_filter_func
        self._fixtures_filter_func = fixtures_filter_func
        self._length = None
        self._shuffle = shuffle
        self._expanded_tests = None

    def __iter__(self):
        if self._expanded_tests is None:  # pragma: no branch
            self._expanded_tests = list(self._tests_generator(self._tests))  # pragma: no cover
        if self._shuffle is True:
            random.shuffle(self._expanded_tests)
        return iter(self._expanded_tests)

    def set_tests_filter(self, filter_func):
        """
        Sets the TestSuites filter function for tests. If the filter function evaluates to False for a given test,
        the test is not yielded.

        Args:
            filter_func (callable): filter function
        """
        assert callable(filter_func)
        self._tests_filter_func = filter_func

    def set_fixtures_filter(self, filter_func):
        """
        Sets the TestSuites filter function for fixtures. If the filter function evaluates to False for a given test,
        the test is not yielded.

        Args:
            filter_func (callable): filter function
        """
        assert callable(filter_func)
        self._fixtures_filter_func = filter_func

    def _tests_generator(self, tests):
        """
        A generator function iterating over tests and applying a filter function if one is set.

        Args:
            tests (list): a list of test functions
        """
        for test in tests:
            if test.is_data_driven():
                for data_driven_test in test.get_data_driven_tests():
                    if self._tests_filter_func is None or self._tests_filter_func(  # pragma: no branch
                                                          data_driven_test, data_driven_test.get_tags()):
                        yield data_driven_test
            else:
                if self._tests_filter_func is None or self._tests_filter_func(test,
                                                                              test.get_tags()):
                    yield test

    def get_fixtures(self):
        """
        Returns a list of all fixtures found by the Discovery.

        Returns:
            List of fixtures
        """

        if self._fixtures_filter_func is not None:
            # filter fixtures by given tags expression
            filtered_fixtures = []

            for fixture in self._fixtures:
                current_tags = self.get_tags_from_fixture(fixture)

                if self._fixtures_filter_func(fixture, current_tags):
                    filtered_fixtures.append(fixture)

            return filtered_fixtures
        else:
            # just return the collected fixtures
            return self._fixtures

    @staticmethod
    def get_tags_from_fixture(fixture):
        tags = set()

        if hasattr(fixture, "_htf_tags"):
            tags |= fixture._htf_tags

        return tags

    def __len__(self):
        """
        The length of a TestSuite is the number of tests it contains.

        Returns:
            Number of tests in the suite
        """
        if self._length is None:
            self._expanded_tests = list(self._tests_generator(self._tests))
            self._length = len(self._expanded_tests)
        return self._length
